"""`Activity streams vocabulary` verbs definitions."""

import sys
from typing import Dict, Optional

from ...base.verbs import BaseXapiVerb
from ...constants import LANG_EN_US_DISPLAY

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class JoinVerb(BaseXapiVerb):
    """Pydantic model for join verb.

    Attributes:
        id (str): Consists of the value `http://activitystrea.ms/join`.
        display (dict): Consists of the dictionary `{"en-US": "joined"}`.
    """

    id: Literal["http://activitystrea.ms/join"] = "http://activitystrea.ms/join"
    display: Optional[Dict[Literal[LANG_EN_US_DISPLAY], Literal["joined"]]]


class LeaveVerb(BaseXapiVerb):
    """Pydantic model for leave `verb`.

    Attributes:
        id (str): Consists of the value `http://activitystrea.ms/leave`.
        display (dict): Consists of the dictionary `{"en-US": "left"}`.
    """

    id: Literal["http://activitystrea.ms/leave"] = "http://activitystrea.ms/leave"
    display: Optional[Dict[Literal[LANG_EN_US_DISPLAY], Literal["left"]]]



"""API-specific data models definition.

Allows to be exactly as lax as we want when it comes to exact object shape and
validation.
"""
from typing import Optional, Union
from uuid import UUID

from pydantic import AnyUrl, BaseModel, Extra

from ..models.xapi.base.agents import BaseXapiAgent
from ..models.xapi.base.groups import BaseXapiGroup


class ErrorDetail(BaseModel):
    """Pydantic model for errors raised detail.

    Type for return value for errors raised in API endpoints.
    Useful for OpenAPI documentation generation.
    """

    detail: str


class BaseModelWithLaxConfig(BaseModel):
    """Pydantic base model with lax configuration.

    Common base lax model to perform light input validation as
    we receive statements through the API.
    """

    class Config:
        """Enable extra properties.

        Useful for not having to perform comprehensive validation.
        """

        extra = Extra.allow


class LaxObjectField(BaseModelWithLaxConfig):
    """Pydantic model for lax `object` field.

    Lightest definition of an object field compliant to the specification.
    """

    id: AnyUrl


class LaxVerbField(BaseModelWithLaxConfig):
    """Pydantic model for lax `verb` field.

    Lightest definition of a verb field compliant to the specification.
    """

    id: AnyUrl


class LaxStatement(BaseModelWithLaxConfig):
    """Pydantic model for lax statement.

    It accepts without validating all fields beyond the bare minimum required to
    qualify an object as an XAPI statement.
    """

    actor: Union[BaseXapiAgent, BaseXapiGroup]
    id: Optional[UUID]
    object: LaxObjectField
    verb: LaxVerbField



"""`AcrossX Profile` activity types definitions."""

import sys

from ...base.unnested_objects import BaseXapiActivity, BaseXapiActivityDefinition

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


# Message


class MessageActivityDefinition(BaseXapiActivityDefinition):
    """Pydantic model for message `Activity` type `definition` property.

    Attributes:
        type (str): Consists of the value
            `https://w3id.org/xapi/acrossx/activities/message`.
    """

    type: Literal[
        "https://w3id.org/xapi/acrossx/activities/message"
    ] = "https://w3id.org/xapi/acrossx/activities/message"


class MessageActivity(BaseXapiActivity):
    """Pydantic model for message `Activity` type.

    Attributes:
        definition (dict): see MessageActivityDefinition.
    """

    definition: MessageActivityDefinition = MessageActivityDefinition()


# Webpage


class WebpageActivityDefinition(BaseXapiActivityDefinition):
    """Pydantic model for webpage `Activity` type `definition` property.

    Attributes:
        type (str): Consists of the value
            `https://w3id.org/xapi/acrossx/activities/webpage`.
    """

    type: Literal["https://w3id.org/xapi/acrossx/activities/webpage"]


class WebpageActivity(BaseXapiActivity):
    """Pydantic model for webpage `Activity` type.

    Attributes:
        definition (dict): see WebpageActivityDefinition.
    """

    definition: WebpageActivityDefinition



"""ClickHouse LRS backend for Ralph."""

import logging
from typing import Generator, Iterator, List, Optional

from ralph.backends.data.clickhouse import (
    ClickHouseDataBackend,
    ClickHouseDataBackendSettings,
)
from ralph.backends.lrs.base import (
    AgentParameters,
    BaseLRSBackend,
    BaseLRSBackendSettings,
    RalphStatementsQuery,
    StatementQueryResult,
)
from ralph.conf import BaseSettingsConfig
from ralph.exceptions import BackendException, BackendParameterException

logger = logging.getLogger(__name__)


class ClickHouseLRSBackendSettings(
    BaseLRSBackendSettings, ClickHouseDataBackendSettings
):
    """ClickHouse LRS backend default configuration.

    Attributes:
        IDS_CHUNK_SIZE (int): The chunk size for querying by ids.
    """

    class Config(BaseSettingsConfig):
        """Pydantic Configuration."""

        env_prefix = "RALPH_BACKENDS__LRS__CLICKHOUSE__"

    IDS_CHUNK_SIZE: int = 10000


class ClickHouseLRSBackend(
    BaseLRSBackend[ClickHouseLRSBackendSettings], ClickHouseDataBackend
):
    """ClickHouse LRS backend implementation."""

    def query_statements(
        self, params: RalphStatementsQuery, target: Optional[str] = None
    ) -> StatementQueryResult:
        """Return the statements query payload using xAPI parameters."""
        ch_params = params.dict(exclude_none=True)
        where = []

        if params.statement_id:
            where.append("event_id = {statementId:UUID}")

        self._add_agent_filters(ch_params, where, params.agent, "actor")
        ch_params.pop("agent", None)

        self._add_agent_filters(ch_params, where, params.authority, "authority")
        ch_params.pop("authority", None)

        if params.verb:
            where.append("JSONExtractString(event, 'verb', 'id') = {verb:String}")

        if params.activity:
            where.append(
                "JSONExtractString(event, 'object', 'objectType') = 'Activity'"
            )
            where.append("JSONExtractString(event, 'object', 'id') = {activity:String}")

        if params.since:
            where.append("emission_time > {since:DateTime64(6)}")

        if params.until:
            where.append("emission_time <= {until:DateTime64(6)}")

        if params.search_after:
            search_order = ">" if params.ascending else "<"

            where.append(
                f"(emission_time {search_order} "
                "{search_after:DateTime64(6)}"
                " OR "
                "(emission_time = {search_after:DateTime64(6)}"
                " AND "
                f"event_id {search_order} "
                "{pit_id:UUID}"
                "))"
            )

        sort_order = "ASCENDING" if params.ascending else "DESCENDING"
        order_by = f"emission_time {sort_order}, event_id {sort_order}"

        query = self.query_class(
            select=["event_id", "emission_time", "event"],
            where=where,
            parameters=ch_params,
            limit=params.limit,
            sort=order_by,
        )
        try:
            clickhouse_response = list(
                self.read(
                    query=query,
                    target=target,
                    ignore_errors=True,
                )
            )
        except (BackendException, BackendParameterException) as error:
            logger.error("Failed to read from ClickHouse")
            raise error

        new_search_after = None
        new_pit_id = None

        if clickhouse_response:
            # Our search after string is a combination of event timestamp and
            # event id, so that we can avoid losing events when they have the
            # same timestamp, and also avoid sending the same event twice.
            new_search_after = clickhouse_response[-1]["emission_time"].isoformat()
            new_pit_id = str(clickhouse_response[-1]["event_id"])

        return StatementQueryResult(
            statements=[document["event"] for document in clickhouse_response],
            search_after=new_search_after,
            pit_id=new_pit_id,
        )

    def query_statements_by_ids(
        self, ids: List[str], target: Optional[str] = None
    ) -> Iterator[dict]:
        """Yield statements with matching ids from the backend."""

        def chunk_id_list(chunk_size: int = self.settings.IDS_CHUNK_SIZE) -> Generator:
            for i in range(0, len(ids), chunk_size):
                yield ids[i : i + chunk_size]

        query = self.query_class(
            select="event",
            where="event_id IN ({ids:Array(String)})",
            parameters={"ids": ["1"]},
            column_oriented=True,
        )
        try:
            for chunk_ids in chunk_id_list():
                query.parameters["ids"] = chunk_ids
                ch_response = self.read(
                    query=query,
                    target=target,
                    ignore_errors=True,
                )
                yield from (document["event"] for document in ch_response)
        except (BackendException, BackendParameterException) as error:
            msg = "Failed to read from ClickHouse"
            logger.error(msg)
            raise error

    @staticmethod
    def _add_agent_filters(
        ch_params: dict,
        where: list,
        agent_params: AgentParameters,
        target_field: str,
    ) -> None:
        """Add filters relative to agents to `where`."""
        if not agent_params:
            return

        if not isinstance(agent_params, dict):
            agent_params = agent_params.dict()

        if agent_params.get("mbox"):
            ch_params[f"{target_field}__mbox"] = agent_params.get("mbox")
            where.append(
                f"JSONExtractString(event, '{target_field}', 'mbox') = "
                f"{{{target_field}__mbox:String}}"
            )
        elif agent_params.get("mbox_sha1sum"):
            ch_params[f"{target_field}__mbox_sha1sum"] = agent_params.get(
                "mbox_sha1sum"
            )
            where.append(
                f"JSONExtractString(event, '{target_field}', 'mbox_sha1sum') = "
                f"{{{target_field}__mbox_sha1sum:String}}"
            )
        elif agent_params.get("openid"):
            ch_params[f"{target_field}__openid"] = agent_params.get("openid")
            where.append(
                f"JSONExtractString(event, '{target_field}', 'openid') = "
                f"{{{target_field}__openid:String}}"
            )
        elif agent_params.get("account__name"):
            ch_params[f"{target_field}__account__name"] = agent_params.get(
                "account__name"
            )
            where.append(
                f"JSONExtractString(event, '{target_field}', 'account', 'name') = "
                f"{{{target_field}__account__name:String}}"
            )
            ch_params[f"{target_field}__account__home_page"] = agent_params.get(
                "account__home_page"
            )
            where.append(
                f"JSONExtractString(event, '{target_field}', 'account', 'homePage') = "
                f"{{{target_field}__account__home_page:String}}"
            )



"""LMS xAPI events context fields definitions."""

import sys
from datetime import datetime
from typing import List, Optional, Union
from uuid import UUID

from pydantic import Field, NonNegativeFloat, PositiveInt, condecimal, validator

from ..base.contexts import BaseXapiContext, BaseXapiContextContextActivities
from ..base.unnested_objects import BaseXapiActivity
from ..concepts.activity_types.scorm_profile import ProfileActivity
from ..concepts.constants.cmi5_profile import CONTEXT_EXTENSION_SESSION_ID
from ..concepts.constants.lms import (
    CONTEXT_EXTENSION_ENDING_DATE,
    CONTEXT_EXTENSION_ROLE,
    CONTEXT_EXTENSION_STARTING_DATE,
)
from ..concepts.constants.video import (
    CONTEXT_EXTENSION_LENGTH,
    CONTEXT_EXTENSION_QUALITY,
)
from ..config import BaseExtensionModelWithConfig

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class LMSProfileActivity(ProfileActivity):
    """Pydantic model for LMS `profile` activity type.

    Attributes:
        id (str): Consists of the value `https://w3id.org/xapi/lms`.
    """

    id: Literal["https://w3id.org/xapi/lms"] = "https://w3id.org/xapi/lms"


class LMSContextContextActivities(BaseXapiContextContextActivities):
    """Pydantic model for LMS `context`.`contextActivities` property.

    Attributes:
        category (dict or list): see LMSProfileActivity.
    """

    category: Union[
        LMSProfileActivity, List[Union[LMSProfileActivity, BaseXapiActivity]]
    ]

    @validator("category")
    @classmethod
    def check_presence_of_profile_activity_category(
        cls,
        value: Union[
            LMSProfileActivity, List[Union[LMSProfileActivity, BaseXapiActivity]]
        ],
    ) -> Union[LMSProfileActivity, List[Union[LMSProfileActivity, BaseXapiActivity]]]:
        """Check that the category list contains a `LMSProfileActivity`."""
        if isinstance(value, LMSProfileActivity):
            return value
        for activity in value:
            if isinstance(activity, LMSProfileActivity):
                return value
        raise ValueError(
            "The `context.contextActivities.category` field should contain at least "
            "one valid `LMSProfileActivity`"
        )


class LMSContext(BaseXapiContext):
    """Pydantic model for LMS base `context` property.

    Attributes:
        contextActivities (dict): see LMSContextContextActivities.
    """

    contextActivities: LMSContextContextActivities


# Registered to a course/Unregistered to a course


class LMSRegistrationContextExtensions(BaseExtensionModelWithConfig):
    """Pydantic model for LMS registration `context`.`extensions` property.

    Attributes:
        starting_date (datetime): Starting date of the activity, formatted according to
            the normal format of ISO 8601.
        ending_date (datetime): Ending date of the activity, formatted according to the
            normal format of ISO 8601.
        role (str): Role of the actor. For example: `admin`, `manager`, `teacher`,
            `guest`, `learner` or `staff`.
    """

    starting_date: Optional[datetime] = Field(alias=CONTEXT_EXTENSION_STARTING_DATE)
    ending_date: Optional[datetime] = Field(alias=CONTEXT_EXTENSION_ENDING_DATE)
    role: Optional[str] = Field(alias=CONTEXT_EXTENSION_ROLE)


class LMSRegistrationContext(LMSContext):
    """Pydantic model for LMS registration statements `context` property.

    This model is used for `registered to a course` and
    `unregistered to a course` statement templates.

    Attributes:
        extensions (dict): see LMSRegistrationContextExtensions.
    """

    extensions: Optional[LMSRegistrationContextExtensions]


class LMSCommonContextExtensions(BaseExtensionModelWithConfig):
    """Pydantic model for common LMS statements `context`.`extensions` property.

    This model is used for `downloaded a video`, `downloaded a document`,
    `downloaded an audio`, `downloaded a file`,  `uploaded a video`,
    `uploaded a document` and `uploaded an audio`, `uploaded a file`
    statement templates.

    Attributes:
        session_id (uuid): ID of the active session.
    """

    session_id: Optional[UUID] = Field(alias=CONTEXT_EXTENSION_SESSION_ID)


class LMSCommonContext(LMSContext):
    """Pydantic model for LMS common `context` property.

    Attributes:
        extensions (dict): See LMSCommonContextExtensions.
    """

    extensions: Optional[LMSCommonContextExtensions]


class LMSDownloadedVideoContextExtensions(LMSCommonContextExtensions):
    """Pydantic model for LMS downloaded video `context`.`extensions` property.

    Attributes:
        length (float): Length of the video.
        quality (int): Video resolution or quality of the video.
    """

    length: Optional[condecimal(ge=0, decimal_places=3)] = Field(
        alias=CONTEXT_EXTENSION_LENGTH
    )
    quality: Optional[PositiveInt] = Field(alias=CONTEXT_EXTENSION_QUALITY)


class LMSDownloadedVideoContext(LMSContext):
    """Pydantic model for LMS downloaded video `context` property.

    Attributes:
        extensions (dict): See LMSDownloadedVideoContextExtensions.
    """

    extensions: Optional[LMSDownloadedVideoContextExtensions]


class LMSDownloadedAudioContextExtensions(LMSCommonContextExtensions):
    """Pydantic model for LMS downloaded audio `context`.`extensions` property.

    Attributes:
        length (float): Length of the audio.
    """

    length: Optional[NonNegativeFloat] = Field(alias=CONTEXT_EXTENSION_LENGTH)


class LMSDownloadedAudioContext(LMSContext):
    """Pydantic model for LMS downloaded audio `context` property.

    Attributes:
        extensions (dict): See LMSDownloadedAudioContextExtensions.
    """

    extensions: Optional[LMSDownloadedAudioContextExtensions]



"""Tests for test helpers."""

from uuid import uuid4

import pytest

from .helpers import (
    assert_statement_get_responses_are_equivalent,
    mock_statement,
    string_is_date,
    string_is_uuid,
)


def test_helpers_string_is_date():
    """Test that strings representing dates are properly identified."""
    string = "2022-06-22T08:31:38+00:00"
    assert string_is_date(string)

    string = "a2023-06-22T08:31:38+00:00"
    assert not string_is_date(string)


def test_helpers_string_is_uuid():
    """Test that strings representing uuids are properly identified."""
    string = str(uuid4())
    assert string_is_uuid(string)

    string = "not_a_valid_uuid"
    assert not string_is_uuid(string)


@pytest.mark.parametrize(
    "modified_fields,are_equivalent",
    [
        (
            {
                "timestamp": None,
                "version": None,
                "authority": "authority_2",
                "stored": "stored_2",
            },
            True,
        ),
        ({"actor": {"actor_field": "actor_2"}}, False),
        ({"verb": "verb_2"}, False),
        ({"object": "object_2"}, False),
        ({"id": "id_2"}, False),
        ({"result": "result_2"}, False),
        ({"context": "context_2"}, False),
        ({"attachments": "attachments_2"}, False),
        ({"timestamp": "timestamp_2"}, False),
        ({"version": "version_2"}, False),
    ],
)
def test_helpers_assert_statement_get_responses_are_equivalent(
    modified_fields, are_equivalent
):
    """Test the equivalency assertion for get responses.

    Equivalency (term NOT in specification) means that two statements have
    identical values on all fields except `authority`, `stored` and `timestamp`
    (where the value may or may not be identical).
    """
    statement_1 = {
        "actor": {"actor_field": "actor_1"},
        "verb": "verb_1",
        "object": "object_1",
        "id": "id_1",
        "result": "result_1",
        "context": "context_1",
        "attachments": "attachments_1",
        "authority": "authority_1",
        "stored": "stored_1",
        "timestamp": "timestamp_1",
        "version": "version_1",
    }

    # Statement to compare to
    statement_2 = statement_1.copy()
    statement_2.update(modified_fields)
    statement_2 = {key: value for key, value in statement_2.items() if value}

    get_response_1 = {"statements": [statement_1]}
    get_response_2 = {"statements": [statement_2]}

    if are_equivalent:
        assert_statement_get_responses_are_equivalent(get_response_1, get_response_2)
        assert_statement_get_responses_are_equivalent(get_response_2, get_response_1)
    else:
        with pytest.raises(AssertionError, match="are not equivalent"):
            assert_statement_get_responses_are_equivalent(
                get_response_1, get_response_2
            )
        with pytest.raises(AssertionError, match="are not equivalent"):
            assert_statement_get_responses_are_equivalent(
                get_response_2, get_response_1
            )


def test_helpers_assert_statement_get_responses_are_equivalent_length_error():
    """Test that responses with different numbers of statements return an error."""

    statement = {
        "actor": {"actor_field": "actor_1"},
        "verb": "verb_1",
        "object": "object_1",
        "id": "id_1",
        "result": "result_1",
        "context": "context_1",
        "attachments": "attachments_1",
        "authority": "authority_1",
        "stored": "stored_1",
        "timestamp": "timestamp_1",
        "version": "version_1",
    }

    get_response_1 = {"statements": [statement]}
    get_response_2 = {"statements": [statement, statement]}

    with pytest.raises(AssertionError):
        assert_statement_get_responses_are_equivalent(get_response_1, get_response_2)
    with pytest.raises(AssertionError):
        assert_statement_get_responses_are_equivalent(get_response_2, get_response_1)


def test_helpers_mock_statement_no_input():
    """Test that mocked statement have the expected fields."""

    statement = mock_statement()

    assert "id" in statement
    assert "actor" in statement
    assert "verb" in statement
    assert "object" in statement
    assert "timestamp" in statement

    statement = mock_statement(timestamp="")
    assert "timestamp" not in statement


def test_helpers_mock_statement_value_input():
    """Test that mocked statement has the expected fields with value input."""

    reference_statement = {
        "id": str(uuid4()),
        "actor": {
            "account": {
                "homePage": "https://example.com/homepage/",
                "name": str(uuid4()),
            },
            "objectType": "Agent",
        },
        # Note the second statement has no preexisting ID
        "object": {"id": "https://example.com/object-id/1/"},
        "timestamp": "2022-03-15T14:07:51Z",
        "verb": {"id": "https://example.com/verb-id/1/"},
    }

    statement = mock_statement(
        id_=reference_statement["id"],
        actor=reference_statement["actor"],
        verb=reference_statement["verb"],
        object=reference_statement["object"],
        timestamp=reference_statement["timestamp"],
    )

    assert statement == reference_statement


@pytest.mark.parametrize("field", ["actor", "verb", "object", "timestamp"])
@pytest.mark.parametrize("integer", [0, 1, 5])
def test_helpers_mock_statement_integer_input(field, integer):
    """Test that mocked statement fields behave properly with integer input."""

    # Test that fields have same values for same integer input
    statement_1 = mock_statement(**{field: integer})
    statement_2 = mock_statement(**{field: integer})
    assert statement_1[field] == statement_2[field]

    # Test that fields have different values for different integer input
    statement_1 = mock_statement(**{field: integer})
    statement_2 = mock_statement(**{field: integer + 1})
    assert statement_1[field] != statement_2[field]



"""Asynchronous Elasticsearch LRS backend for Ralph."""

import logging
from typing import AsyncIterator, List, Optional

from ralph.backends.data.async_es import AsyncESDataBackend
from ralph.backends.lrs.base import (
    BaseAsyncLRSBackend,
    RalphStatementsQuery,
    StatementQueryResult,
)
from ralph.backends.lrs.es import ESLRSBackend, ESLRSBackendSettings
from ralph.exceptions import BackendException, BackendParameterException

logger = logging.getLogger(__name__)


class AsyncESLRSBackend(BaseAsyncLRSBackend[ESLRSBackendSettings], AsyncESDataBackend):
    """Asynchronous Elasticsearch LRS backend implementation."""

    async def query_statements(
        self, params: RalphStatementsQuery, target: Optional[str] = None
    ) -> StatementQueryResult:
        """Return the statements query payload using xAPI parameters."""
        query = ESLRSBackend.get_query(params=params)
        try:
            statements = [
                document["_source"]
                async for document in self.read(
                    query=query, target=target, chunk_size=params.limit
                )
            ]
        except (BackendException, BackendParameterException) as error:
            logger.error("Failed to read from Elasticsearch")
            raise error

        return StatementQueryResult(
            statements=statements,
            pit_id=query.pit.id,
            search_after="|".join(query.search_after) if query.search_after else "",
        )

    async def query_statements_by_ids(
        self, ids: List[str], target: Optional[str] = None
    ) -> AsyncIterator[dict]:
        """Yield statements with matching ids from the backend."""
        query = self.query_class(query={"terms": {"_id": ids}})
        try:
            async for document in self.read(query=query, target=target):
                yield document["_source"]
        except (BackendException, BackendParameterException) as error:
            logger.error("Failed to read from Elasticsearch")
            raise error



"""Configurations for Ralph."""

import io
import sys
from enum import Enum
from pathlib import Path
from typing import List, Sequence, Tuple, Union

from pydantic import AnyHttpUrl, AnyUrl, BaseModel, BaseSettings, Extra, root_validator

from ralph.exceptions import ConfigurationException

from .utils import import_string

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

try:
    from click import get_app_dir
except ImportError:
    # If we use Ralph as a library and Click is not installed, we consider the
    # application directory to be the current directory. For non-CLI usage, it
    # has no consequences.
    from unittest.mock import Mock

    get_app_dir = Mock(return_value=".")

MODEL_PATH_SEPARATOR = "__"


class BaseSettingsConfig:
    """Pydantic model for BaseSettings Configuration."""

    case_sensitive = True
    env_nested_delimiter = "__"
    env_prefix = "RALPH_"
    extra = "ignore"


class CoreSettings(BaseSettings):
    """Pydantic model for Ralph's core settings."""

    class Config(BaseSettingsConfig):
        """Pydantic Configuration."""

    APP_DIR: Path = get_app_dir("ralph")
    LOCALE_ENCODING: str = getattr(io, "LOCALE_ENCODING", "utf8")


core_settings = CoreSettings()


class CommaSeparatedTuple(str):
    """Pydantic field type validating comma-separated strings or lists/tuples."""

    @classmethod
    def __get_validators__(cls):  # noqa: D105
        def validate(value: Union[str, Sequence[str]]) -> Sequence[str]:
            """Check whether the value is a comma-separated string or a list/tuple."""
            if isinstance(value, (tuple, list)):
                return tuple(value)

            if isinstance(value, str):
                return tuple(value.split(","))

            raise TypeError("Invalid comma-separated list")

        yield validate


class InstantiableSettingsItem(BaseModel):
    """Pydantic model for a settings configuration item that can be instantiated."""

    class Config:  # noqa: D106
        underscore_attrs_are_private = True

    _class_path: str = None

    def get_instance(self, **init_parameters):
        """Return an instance of the settings item class using its `_class_path`."""
        return import_string(self._class_path)(**init_parameters)


class ClientOptions(BaseModel):
    """Pydantic model for additional client options."""

    class Config:  # noqa: D106
        extra = Extra.forbid


class HeadersParameters(BaseModel):
    """Pydantic model for headers parameters."""

    class Config:  # noqa: D106
        extra = Extra.allow


# Active parser Settings.


class ESParserSettings(InstantiableSettingsItem):
    """Pydantic model for Elasticsearch parser configuration settings."""

    _class_path: str = "ralph.parsers.ElasticSearchParser"


class GELFParserSettings(InstantiableSettingsItem):
    """Pydantic model for GELF parser configuration settings."""

    _class_path: str = "ralph.parsers.GELFParser"


class ParserSettings(BaseModel):
    """Pydantic model for parsers configuration settings."""

    GELF: GELFParserSettings = GELFParserSettings()
    ES: ESParserSettings = ESParserSettings()


class XapiForwardingConfigurationSettings(BaseModel):
    """Pydantic model for xAPI forwarding configuration item."""

    class Config:  # noqa: D106
        min_anystr_length = 1

    url: AnyUrl
    is_active: bool
    basic_username: str
    basic_password: str
    max_retries: int
    timeout: float


class AuthBackend(str, Enum):
    """Model for valid authentication methods."""

    BASIC = "basic"
    OIDC = "oidc"


class AuthBackends(Tuple[AuthBackend]):
    """Model representing a tuple of authentication backends."""

    @classmethod
    def __get_validators__(cls):
        """Check whether the value is a comma-separated string or a tuple representing
        an AuthBackend.
        """  # noqa: D205

        def validate(
            auth_backends: Union[
                str, AuthBackend, Tuple[AuthBackend], List[AuthBackend]
            ]
        ) -> Tuple[AuthBackend]:
            """Check whether the value is a comma-separated string or a list/tuple."""
            if isinstance(auth_backends, str):
                return tuple(
                    AuthBackend(value.lower()) for value in auth_backends.split(",")
                )

            if isinstance(auth_backends, AuthBackend):
                return (auth_backends,)

            if isinstance(auth_backends, (tuple, list)):
                return tuple(auth_backends)

            raise TypeError("Invalid comma-separated list")

        yield validate


class Settings(BaseSettings):
    """Pydantic model for Ralph's global environment & configuration settings."""

    class Config(BaseSettingsConfig):
        """Pydantic Configuration."""

        env_file = ".env"
        env_file_encoding = core_settings.LOCALE_ENCODING

    _CORE: CoreSettings = core_settings
    AUTH_FILE: Path = _CORE.APP_DIR / "auth.json"
    AUTH_CACHE_MAX_SIZE = 100
    AUTH_CACHE_TTL = 3600
    CONVERTER_EDX_XAPI_UUID_NAMESPACE: str = None
    EXECUTION_ENVIRONMENT: str = "development"
    HISTORY_FILE: Path = _CORE.APP_DIR / "history.json"
    LOGGING: dict = {
        "version": 1,
        "propagate": True,
        "formatters": {
            "ralph": {
                "format": "%(asctime)-23s %(levelname)-8s %(name)-8s %(message)s"
            },
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "level": "INFO",
                "stream": "ext://sys.stderr",
                "formatter": "ralph",
            },
        },
        "loggers": {
            "ralph": {
                "handlers": ["console"],
                "level": "INFO",
            },
            "swiftclient": {
                "handlers": ["console"],
                "level": "ERROR",
            },
            "uvicorn": {
                "handlers": ["console"],
                "level": "INFO",
            },
        },
    }
    PARSERS: ParserSettings = ParserSettings()
    RUNSERVER_AUTH_BACKENDS: AuthBackends = AuthBackends([AuthBackend.BASIC])
    RUNSERVER_AUTH_OIDC_AUDIENCE: str = None
    RUNSERVER_AUTH_OIDC_ISSUER_URI: AnyHttpUrl = None
    RUNSERVER_BACKEND: Literal[
        "async_es", "async_mongo", "clickhouse", "es", "fs", "mongo"
    ] = "es"
    RUNSERVER_HOST: str = "0.0.0.0"  # noqa: S104
    RUNSERVER_MAX_SEARCH_HITS_COUNT: int = 100
    RUNSERVER_POINT_IN_TIME_KEEP_ALIVE: str = "1m"
    RUNSERVER_PORT: int = 8100
    LRS_RESTRICT_BY_AUTHORITY: bool = False
    LRS_RESTRICT_BY_SCOPES: bool = False
    SENTRY_CLI_TRACES_SAMPLE_RATE: float = 1.0
    SENTRY_DSN: str = None
    SENTRY_IGNORE_HEALTH_CHECKS: bool = False
    SENTRY_LRS_TRACES_SAMPLE_RATE: float = 1.0
    XAPI_FORWARDINGS: List[XapiForwardingConfigurationSettings] = []

    @property
    def APP_DIR(self) -> Path:
        """Return the path to Ralph's configuration directory."""
        return self._CORE.APP_DIR

    @property
    def LOCALE_ENCODING(self) -> str:
        """Return Ralph's default locale encoding."""
        return self._CORE.LOCALE_ENCODING

    @root_validator(allow_reuse=True)
    @classmethod
    def check_restriction_compatibility(cls, values):
        """Raise an error if scopes are being used without authority restriction."""
        if values.get("LRS_RESTRICT_BY_SCOPES") and not values.get(
            "LRS_RESTRICT_BY_AUTHORITY"
        ):
            raise ConfigurationException(
                "LRS_RESTRICT_BY_AUTHORITY must be set to True if using "
                "LRS_RESTRICT_BY_SCOPES=True"
            )
        return values


settings = Settings()



"""OpenID Connect authentication tool for the Ralph API."""

import logging
from functools import lru_cache
from typing import Dict, Optional

import requests
from fastapi import Depends, HTTPException, status
from fastapi.security import HTTPBearer, OpenIdConnect
from jose import ExpiredSignatureError, JWTError, jwt
from jose.exceptions import JWTClaimsError
from pydantic import AnyUrl, BaseModel, Extra
from typing_extensions import Annotated

from ralph.api.auth.user import AuthenticatedUser, UserScopes
from ralph.conf import settings

OPENID_CONFIGURATION_PATH = "/.well-known/openid-configuration"
oauth2_scheme = OpenIdConnect(
    openIdConnectUrl=f"""{settings.RUNSERVER_AUTH_OIDC_ISSUER_URI}
    {OPENID_CONFIGURATION_PATH}""",
    auto_error=False,
)

# API auth logger
logger = logging.getLogger(__name__)


class IDToken(BaseModel):
    """Pydantic model representing the core of an OpenID Connect ID Token.

    ID Tokens are polymorphic and may have many attributes not defined in the
    specification. This model ignores all additional fields.

    Attributes:
        iss (str): Issuer Identifier for the Issuer of the response.
        sub (str): Subject Identifier.
        aud (str): Audience(s) that this ID Token is intended for.
        exp (int): Expiration time on or after which the ID Token MUST NOT be
                   accepted for processing.
        iat (int): Time at which the JWT was issued.
        scope (str): Scope(s) for resource authorization.
        target (str): Target for storing the statements.
    """

    iss: str
    sub: str
    aud: Optional[str]
    exp: int
    iat: int
    scope: Optional[str]
    target: Optional[str]

    class Config:  # noqa: D106
        extra = Extra.ignore


@lru_cache()
def discover_provider(base_url: AnyUrl) -> Dict:
    """Discover the authentication server (or OpenId Provider) configuration."""
    try:
        response = requests.get(f"{base_url}{OPENID_CONFIGURATION_PATH}", timeout=5)
        response.raise_for_status()
        return response.json()
    except requests.exceptions.RequestException as exc:
        logger.error(
            "Unable to discover the authentication server configuration: %s", exc
        )
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Could not validate credentials",
            headers={"WWW-Authenticate": "Bearer"},
        ) from exc


@lru_cache()
def get_public_keys(jwks_uri: AnyUrl) -> Dict:
    """Retrieve the public keys used by the provider server for signing."""
    try:
        response = requests.get(jwks_uri, timeout=5)
        response.raise_for_status()
        return response.json()
    except requests.exceptions.RequestException as exc:
        logger.error(
            (
                "Unable to retrieve the public keys used by the provider server"
                "for signing: %s"
            ),
            exc,
        )
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Could not validate credentials",
            headers={"WWW-Authenticate": "Bearer"},
        ) from exc


def get_oidc_user(
    auth_header: Annotated[Optional[HTTPBearer], Depends(oauth2_scheme)],
) -> AuthenticatedUser:
    """Decode and validate OpenId Connect ID token against issuer in config.

    Args:
        auth_header (str): Authentication header containing the Base64 encoded
            OIDC Token. This is invoked behind the scenes by Depends.
        security_scopes (SecurityScopes): Scopes required to access the endpoint.

    Return:
        AuthenticatedUser (AuthenticatedUser)

    Raises:
        HTTPException
    """
    if auth_header is None or "bearer" not in auth_header.lower():
        logger.debug(
            "Not using OIDC auth. The OpenID Connect authentication mode requires a "
            "Bearer token"
        )
        return None

    id_token = auth_header.split(" ")[-1]
    provider_config = discover_provider(settings.RUNSERVER_AUTH_OIDC_ISSUER_URI)
    key = get_public_keys(provider_config["jwks_uri"])
    algorithms = provider_config["id_token_signing_alg_values_supported"]
    audience = settings.RUNSERVER_AUTH_OIDC_AUDIENCE
    options = {
        "verify_signature": True,
        "verify_aud": bool(audience),
        "verify_exp": True,
    }
    try:
        decoded_token = jwt.decode(
            token=id_token,
            key=key,
            algorithms=algorithms,
            options=options,
            audience=audience,
        )
    except (ExpiredSignatureError, JWTError, JWTClaimsError) as exc:
        logger.error("Unable to decode the ID token: %s", exc)
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Could not validate credentials",
            headers={"WWW-Authenticate": "Bearer"},
        ) from exc

    id_token = IDToken.parse_obj(decoded_token)

    user = AuthenticatedUser(
        agent={"openid": f"{id_token.iss}/{id_token.sub}"},
        scopes=UserScopes(id_token.scope.split(" ") if id_token.scope else []),
        target=id_token.target,
    )
    return user



"""`AcrossX Profile` verbs definitions."""

import sys
from typing import Dict, Optional

from ...base.verbs import BaseXapiVerb
from ...constants import LANG_EN_US_DISPLAY

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class PostedVerb(BaseXapiVerb):
    """Pydantic model for posted `verb`.

    Attributes:
        id (str): Consists of the value `https://w3id.org/xapi/acrossx/verbs/posted`.
        display (dict): Consists of the dictionary `{"en-US": "posted"}`.
    """

    id: Literal[
        "https://w3id.org/xapi/acrossx/verbs/posted"
    ] = "https://w3id.org/xapi/acrossx/verbs/posted"
    display: Optional[Dict[Literal[LANG_EN_US_DISPLAY], Literal["posted"]]]



"""Base xAPI `Agent` definitions."""

import sys
from abc import ABC
from typing import Optional, Union

from pydantic import StrictStr

from ..config import BaseModelWithConfig
from .common import IRI
from .ifi import (
    BaseXapiAccountIFI,
    BaseXapiMboxIFI,
    BaseXapiMboxSha1SumIFI,
    BaseXapiOpenIdIFI,
)

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class BaseXapiAgentAccount(BaseModelWithConfig):
    """Pydantic model for `Agent` type `account` property.

    Attributes:
        homePage (IRI): Consists of the home page of the account's service provider.
        name (str): Consists of the unique id or name of the Actor's account.
    """

    homePage: IRI
    name: StrictStr


class BaseXapiAgentCommonProperties(BaseModelWithConfig, ABC):
    """Pydantic model for core `Agent` type property.

    It defines who performed the action.

    Attributes:
        objectType (str): Consists of the value `Agent`.
        name (str): Consists of the full name of the Agent.
    """

    objectType: Optional[Literal["Agent"]]
    name: Optional[StrictStr]


class BaseXapiAgentWithMbox(BaseXapiAgentCommonProperties, BaseXapiMboxIFI):
    """Pydantic model for `Agent` type property.

    It is defined for agent type with a mailto IFI.
    """


class BaseXapiAgentWithMboxSha1Sum(
    BaseXapiAgentCommonProperties, BaseXapiMboxSha1SumIFI
):
    """Pydantic model for `Agent` type property.

    It is defined for agent type with a hash IFI.
    """


class BaseXapiAgentWithOpenId(BaseXapiAgentCommonProperties, BaseXapiOpenIdIFI):
    """Pydantic model for `Agent` type property.

    It is defined for agent type with an openID IFI.
    """


class BaseXapiAgentWithAccount(BaseXapiAgentCommonProperties, BaseXapiAccountIFI):
    """Pydantic model for `Agent` type property.

    It is defined for agent type with an account IFI.
    """


BaseXapiAgent = Union[
    BaseXapiAgentWithMbox,
    BaseXapiAgentWithMboxSha1Sum,
    BaseXapiAgentWithOpenId,
    BaseXapiAgentWithAccount,
]



"""Constants for `Scorm Profile` xAPI profile."""

# Context extensions
CONTEXT_EXTENSION_COURSE_ID = "http://adlnet.gov/expapi/activities/course"
CONTEXT_EXTENSION_MODULE_ID = "http://adlnet.gov/expapi/activities/module"



"""edX to xAPI conversion sets."""

# flake8: noqa

from .enrollment import (
    EdxCourseEnrollmentActivatedToLMSRegisteredCourse,
    EdxCourseEnrollmentDeactivatedToLMSUnregisteredCourse,
)
from .navigational import UIPageCloseToPageTerminated
from .server import ServerEventToPageViewed
from .video import (
    UILoadVideoToVideoInitialized,
    UIPauseVideoToVideoPaused,
    UIPlayVideoToVideoPlayed,
    UISeekVideoToVideoSeeked,
    UIStopVideoToVideoTerminated,
)



"""Common for xAPI base definitions."""

from typing import Dict, Generator, Type

from langcodes import tag_is_valid
from pydantic import StrictStr, validate_email
from rfc3987 import parse


class IRI(str):
    """Pydantic custom data type validating RFC 3987 IRIs."""

    @classmethod
    def __get_validators__(cls) -> Generator:  # noqa: D105
        def validate(iri: str) -> Type["IRI"]:
            """Check whether the provided IRI is a valid RFC 3987 IRI."""
            parse(iri, rule="IRI")
            return cls(iri)

        yield validate


class LanguageTag(str):
    """Pydantic custom data type validating RFC 5646 Language tags."""

    @classmethod
    def __get_validators__(cls) -> Generator:  # noqa: D105
        def validate(tag: str) -> Type["LanguageTag"]:
            """Check whether the provided tag is a valid RFC 5646 Language tag."""
            if not tag_is_valid(tag):
                raise TypeError("Invalid RFC 5646 Language tag")
            return cls(tag)

        yield validate


LanguageMap = Dict[LanguageTag, StrictStr]


class MailtoEmail(str):
    """Pydantic custom data type validating `mailto:email` format."""

    @classmethod
    def __get_validators__(cls) -> Generator:  # noqa: D105
        def validate(mailto: str) -> Type["MailtoEmail"]:
            """Check whether the provided value follows the `mailto:email` format."""
            if not mailto.startswith("mailto:"):
                raise TypeError("Invalid `mailto:email` value")
            valid = validate_email(mailto[7:])
            return cls(f"mailto:{valid[1]}")

        yield validate



"""`Scorm Profile` activity types definitions."""

import sys

from ...base.unnested_objects import BaseXapiActivity, BaseXapiActivityDefinition

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


# Document


class DocumentActivityDefinition(BaseXapiActivityDefinition):
    """Pydantic model for document `Activity` type `definition` property.

    Attributes:
        type (str): Consists of the value
            `http://id.tincanapi.com/activitytype/document`.
    """

    type: Literal["http://id.tincanapi.com/activitytype/document"]


class DocumentActivity(BaseXapiActivity):
    """Pydantic model for document `Activity` type.

    Attributes:
        definition (dict): see DocumentActivityDefinition.
    """

    definition: DocumentActivityDefinition



"""Tests for Ralph flavors dependencies"""

import importlib
import re
import sys

import pytest
from click.testing import CliRunner

from ralph import cli


def test_dependencies_ralph_command_requires_click(monkeypatch):
    """Test Click module installation while executing the ralph command."""

    monkeypatch.setitem(sys.modules, "click", None)

    # Force ralph.cli reload now that click is considered as missing
    if "ralph.cli" in sys.modules:
        del sys.modules["ralph.cli"]

    with pytest.raises(
        ModuleNotFoundError,
        match=re.escape(
            "You need to install 'cli' optional dependencies to use the ralph "
            "command: pip install ralph-malph[cli]"
        ),
    ):
        importlib.import_module("ralph.cli")


def test_dependencies_runserver_subcommand_requires_uvicorn(monkeypatch):
    """Test Uvicorn module installation while executing the runserver sub command."""

    monkeypatch.delattr(cli, "uvicorn")
    monkeypatch.setattr(cli, "configure_logging", lambda: None)
    runner = CliRunner()
    result = runner.invoke(cli.cli, "runserver -b es".split())
    assert isinstance(result.exception, ModuleNotFoundError)
    assert str(result.exception) == (
        "You need to install 'lrs' optional dependencies to use the runserver "
        "command: pip install ralph-malph[lrs]"
    )



"""`Video` activity types definitions."""

import sys

from ...base.unnested_objects import BaseXapiActivity, BaseXapiActivityDefinition

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


# Video


class VideoActivityDefinition(BaseXapiActivityDefinition):
    """Pydantic model for video `Activity` type `definition` property.

    Attributes:
        type (str): Consists of the value
            `https://w3id.org/xapi/video/activity-type/video`.
    """

    type: Literal[
        "https://w3id.org/xapi/video/activity-type/video"
    ] = "https://w3id.org/xapi/video/activity-type/video"


class VideoActivity(BaseXapiActivity):
    """Pydantic model for video `Activity` type.

    Attributes:
        definition (dict): See VideoActivityDefinition.
    """

    definition: VideoActivityDefinition = VideoActivityDefinition()



"""Constants for `LMS` xAPI profile."""

# Context extension IDs
CONTEXT_EXTENSION_STARTING_DATE = "https://w3id.org/xapi/lms/extensions/starting-date"
CONTEXT_EXTENSION_ENDING_DATE = "https://w3id.org/xapi/lms/extensions/ending-date"
CONTEXT_EXTENSION_ROLE = "https://w3id.org/xapi/lms/extensions/role"



"""Peer instruction event field definition."""

from pydantic import constr

from ...base import AbstractBaseEventField


class PeerInstructionEventField(AbstractBaseEventField):
    """Pydantic model for peer instruction `event` field.

    Attributes:
        answer (int): Consists of the index assigned to the answer choice selected by
            the learner.
        rationale (str): Consists of the text entered by the learner to explain why
            they selected that answer choice.
        truncated (bool): `True` only if the rationale was longer than 12,500
            characters, which is the maximum included in the event.
    """

    answer: int
    rationale: constr(max_length=12500)
    truncated: bool



"""Problem interaction events model event fields definitions."""

import sys
from datetime import datetime
from typing import Dict, List, Optional, Union

from pydantic import constr

from ...base import AbstractBaseEventField, BaseModelWithConfig

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class QueueState(BaseModelWithConfig):
    """Pydantic model for problem interaction `event`.`correct_map`.`queuestate` field.

    Attributes:
        key (str): Consists of a secret string.
        time (str): Consists of a string dump of a DateTime object in the format
            '%Y%m%d%H%M%S'.
    """

    key: str
    time: datetime


class CorrectMap(BaseModelWithConfig):
    """Pydantic model for problem interaction `event`.`correct_map` field.

    Attributes:
        answervariable (str): Consists of the variable chosen in answer in the case of
            optionresponse provided with variables.
        correctness (str): Consists either of the `correct` or `incorrect` value.
        hint (str): Consists of optional hint.
        hint_mode (str): Consists either of the value `on_request` or `always` value.
        msg (str): Consists of extra message response.
        npoints (int): Consists of awarded points.
        queuestate (json): see QueueStateField.
    """

    answervariable: Union[Literal[None], None, str]
    correctness: Union[Literal["correct"], Literal["incorrect"]]
    hint: Optional[str]
    hintmode: Optional[Union[Literal["on_request"], Literal["always"]]]
    msg: str
    npoints: Optional[int]
    queuestate: Optional[QueueState]


class State(BaseModelWithConfig):
    """Pydantic model for problem interaction `event`.`state` field.

    Attributes:
        correct_map (dict): see CorrectMapSubFields.
        done (bool): `True` if the problem is answered, else `False`.
        input_state (dict): Consists of the state field given before answering.
        seed (int): Consists of the seed element for the current state.
        student_answers (dict): Consists of the answer(s) given by the user.
    """

    correct_map: Dict[
        constr(regex=r"^[a-f0-9]{32}_[0-9]_[0-9]$"),
        CorrectMap,
    ]
    done: Optional[bool]
    input_state: dict
    seed: int
    student_answers: dict


class SubmissionAnswerField(BaseModelWithConfig):
    """Pydantic model for `problem_check`.`event`.`submission` field.

    Attributes:
        answer (str, list): Consists of the answer string or a list of the answer
            strings if multiple choices are allowed.
        correct (bool): `True` if the `answer` value is correct, else `False`.
        input_type (str): Consists of the type of value that the student supplies for
            the `response_type`.
        question (str): Consists of the question text.
        response_type (str): Consists of the type of problem.
        variant (str): Consists of the unique ID of the variant that was presented to
            this user.
    """

    answer: Union[str, List[str]]
    correct: bool
    input_type: str
    question: str
    response_type: str
    variant: str


class EdxProblemHintDemandhintDisplayedEventField(AbstractBaseEventField):
    """Pydantic model for `edx.problem.hint.demandhint_displayed`.`event` field.

    Attributes:
        hint_index (int): Consists of the identifier for the hint that was displayed to
            the user.
        hint_len (int): Consists of the total number of hints defined for this problem.
        hint_text (str): Consists of the text of the hint that was displayed to the
            user.
        module_id (str): Consists of the identifier for the problem component for which
            the user requested the hint.
    """

    hint_index: int
    hint_len: int
    hint_text: str
    module_id: str


class EdxProblemHintFeedbackDisplayedEventField(AbstractBaseEventField):
    """Pydantic model for `edx.problem.hint.feedback_displayed`.`event` field.

    Attributes:
        choice_all (list): Lists all the answer choices for problems with multiple
            possible answers defined.
        correctness (bool): `True` if the `student_answer` value is correct, else
            `False`.
        hint_label (str): Consists of the feedback message given for the answer
            correctness.
        hints (list): Consists of a text member field with the given feedback string.
        module_id (str): Consists of the identifier for the problem component for which
            the user received the feedback.
        problem_part_id (str): Consists of the specific problem for which the user
            received feedback.
        question_type (str): Consists of the XML tag that identifies the problem type.
        student_answer (list): Consists of the answer value(s) selected or supplied by
            the user.
        trigger_type (str): Identifies the type of feedback obtained by the
            `student_answer` response. Consists either of `single` or `compound` value.
    """

    choice_all: Optional[List[str]]
    correctness: bool
    hint_label: str
    hints: List[dict]
    module_id: str
    problem_part_id: str
    question_type: Union[
        Literal["stringresponse"],
        Literal["choiceresponse"],
        Literal["multiplechoiceresponse"],
        Literal["numericalresponse"],
        Literal["optionresponse"],
    ]
    student_answer: List[str]
    trigger_type: Union[Literal["single"], Literal["compound"]]


class ProblemCheckEventField(AbstractBaseEventField):
    """Pydantic model for `problem_check`.`event` field.

    Attributes:
        answers (dict): Consists of a dictionary of problem ID and the corresponding
            internal answer identifier for each problem.
        attempts (int): Consists of the number of times the user attempted to answer
            the problem.
        correct_map (dict): Consists of the evaluation data for each answer.
        grade (int): Consists of the current grade value.
        max_grade (int): Consists of the maximum possible grade value.
        problem_id (str): Consists of the ID of the problem that was checked.
        state (json): Consists of the current problem state.
        submission (dict): Consists of a dictionary of data about the given answer.
        success (str): Consists of either the `correct` or `incorrect` value.
    """

    answers: Dict[
        constr(regex=r"^[a-f0-9]{32}_[0-9]_[0-9]$"),
        Union[List[str], str],
    ]
    attempts: int
    correct_map: Dict[
        constr(regex=r"^[a-f0-9]{32}_[0-9]_[0-9]$"),
        CorrectMap,
    ]
    grade: int
    max_grade: int
    problem_id: constr(
        regex=r"^block-v1:[^\/+]+(\/|\+)[^\/+]+(\/|\+)[^\/?]+"
        r"type@problem\+block@[a-f0-9]{32}$"
    )
    state: State
    submission: Dict[
        constr(regex=r"^[a-f0-9]{32}_[0-9]_[0-9]$"),
        SubmissionAnswerField,
    ]
    success: Union[Literal["correct"], Literal["incorrect"]]


class ProblemCheckFailEventField(AbstractBaseEventField):
    """Pydantic model for `problem_check_fail`.`event` field.

    Attributes:
        answers (dict): Consists of a dictionary of problem ID and the internal answer
            identifier for each problem.
        failure (str): Consists either of the `closed` or `unreset` value.
        problem_id (str): Consists of the ID of the problem that was checked.
        state (dict): Consists of the current problem state.
    """

    answers: Dict[
        constr(regex=r"^[a-f0-9]{32}_[0-9]_[0-9]$"),
        Union[List[str], str],
    ]
    failure: Union[Literal["closed"], Literal["unreset"]]
    problem_id: constr(
        regex=r"^block-v1:[^\/+]+(\/|\+)[^\/+]+(\/|\+)[^\/?]+"
        r"type@problem\+block@[a-f0-9]{32}$"
    )
    state: State


class ProblemRescoreEventField(AbstractBaseEventField):
    """Pydantic model for `problem_rescore`.`event` field.

    Attributes:
        attempts (int): Consists of the number of attempts of rescoring.
        correct_map (json): see CorrectMapSubFields.
        new_score (int): Consists of the new score obtained after rescoring.
        new_total (int): Consists of the new total summed after rescoring.
        orig_score (int): Consists of the original scored before rescoring.
        problem_id (str): Consists of the ID of the problem being rescored.
        state (json): see StateField.
        success (str): Consists either of the `correct` or `incorrect` value.
    """

    attempts: int
    correct_map: CorrectMap
    new_score: int
    new_total: int
    orig_score: int
    orig_total: int
    problem_id: constr(
        regex=r"^block-v1:[^\/+]+(\/|\+)[^\/+]+(\/|\+)[^\/?]+"
        r"type@problem\+block@[a-f0-9]{32}$"
    )
    state: State
    success: Union[Literal["correct"], Literal["incorrect"]]


class ProblemRescoreFailEventField(AbstractBaseEventField):
    """Pydantic model for `problem_rescore_fail`.`event` field.

    Attributes:
        failure (str): Consists either of the `closed` or `unreset` value.
        problem_id (str): Consists of the ID of the problem being checked.
        state (json): see StateField.
    """

    failure: Union[Literal["closed"], Literal["unreset"]]
    problem_id: constr(
        regex=r"^block-v1:[^\/+]+(\/|\+)[^\/+]+(\/|\+)[^\/?]+"
        r"type@problem\+block@[a-f0-9]{32}$"
    )
    state: State


class UIProblemResetEventField(AbstractBaseEventField):
    """Pydantic model for `problem_reset`.`event` field.

    Attributes:
        answers (str, list): Consists of the answer string or a list of the answer
            strings if multiple choices are allowed.
    """

    answers: Union[str, List[str]]


class UIProblemShowEventField(AbstractBaseEventField):
    """Pydantic model for `problem_show`.`event` field.

    Attributes:
        problem (str): Consists of the optional name value that the course creators
            supply or the system-generated hash code for the problem being shown.
    """

    problem: str


class ResetProblemEventField(AbstractBaseEventField):
    """Pydantic model for `reset_problem`.`event` field.

    Attributes:
        new_state (json): see StateField.
        old_state (json): see StateField.
        problem_id (str): Consists of the ID of the problem being reset.
    """

    new_state: State
    old_state: State
    problem_id: constr(
        regex=r"^block-v1:[^\/+]+(\/|\+)[^\/+]+(\/|\+)[^\/?]+"
        r"type@problem\+block@[a-f0-9]{32}$"
    )


class ResetProblemFailEventField(AbstractBaseEventField):
    """Pydantic model for `reset_problem_fail`.`event` field.

    Attributes:
        failure (str): Consists either of `closed` or `not_done` value.
        old_state (json): see StateField.
        problem_id (str): Consists of the ID of the problem being reset.
    """

    failure: Union[Literal["closed"], Literal["not_done"]]
    old_state: State
    problem_id: constr(
        regex=r"^block-v1:[^\/+]+(\/|\+)[^\/+]+(\/|\+)[^\/?]+"
        r"type@problem\+block@[a-f0-9]{32}$"
    )


class SaveProblemFailEventField(AbstractBaseEventField):
    """Pydantic model for `save_problem_fail`.`event` field.

    Attributes:
        answers (dict): Consists of a dict of the answer string or a list or a dict of
            the answer strings if multiple choices are allowed.
        failure (str): Consists either of `closed` or `done` value.
        problem_id (str): Consists of the ID of the problem being saved.
        state (json): see StateField.
    """

    answers: Dict[str, Union[int, str, list, dict]]
    failure: Union[Literal["closed"], Literal["done"]]
    problem_id: constr(
        regex=r"^block-v1:[^\/+]+(\/|\+)[^\/+]+(\/|\+)[^\/?]+"
        r"type@problem\+block@[a-f0-9]{32}$"
    )
    state: State


class SaveProblemSuccessEventField(AbstractBaseEventField):
    """Pydantic model for `save_problem_success`.`event` field.

    Attributes:
        answers (dict): Consists of a dict of the answer string or a list or a dict of
            the answer strings if multiple choices are allowed.
        problem_id (str): Consists of the ID of the problem being saved.
        state (json): see StateField.
    """

    answers: Dict[str, Union[int, str, list, dict]]
    problem_id: constr(
        regex=r"^block-v1:[^\/+]+(\/|\+)[^\/+]+(\/|\+)[^\/?]+"
        r"type@problem\+block@[a-f0-9]{32}$"
    )
    state: State


class ShowAnswerEventField(AbstractBaseEventField):
    """Pydantic model for `show_answer`.`event` field.

    Attributes:
        problem_id (str): Consists of the ID of the problem being shown.
    """

    problem_id: constr(
        regex=r"^block-v1:[^\/+]+(\/|\+)[^\/+]+(\/|\+)[^\/?]+"
        r"type@problem\+block@[a-f0-9]{32}$"
    )



"""Peer instruction events model definitions."""

import sys
from typing import Union

from pydantic import Json

from ralph.models.selector import selector

from ..server import BaseServerModel
from .fields.events import PeerInstructionEventField

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class PeerInstructionAccessed(BaseServerModel):
    """Pydantic model for `ubc.peer_instruction.accessed` statement.

    The server emits this event when a peer instruction question and its set of answer
    choices is shown to a learner.

    Attributes:
        event_type (str): Consists of the value `ubc.peer_instruction.accessed`.
        name (str): Consists of the value `ubc.peer_instruction.accessed`.
    """

    __selector__ = selector(
        event_source="server", event_type="ubc.peer_instruction.accessed"
    )

    event_type: Literal["ubc.peer_instruction.accessed"]
    name: Literal["ubc.peer_instruction.accessed"]


class PeerInstructionOriginalSubmitted(BaseServerModel):
    """Pydantic model for `ubc.peer_instruction.original_submitted` statement.

    The server emits this event when learners submit their initial responses. These
    events record the answer choice the learner selected and the explanation given
    for why that selection was made.

    Attributes:
        event (int): See PeerInstructionEventField.
        event_type (str): Consists of the value
            `ubc.peer_instruction.original_submitted`.
        name (str): Consists of the value `ubc.peer_instruction.original_submitted`.
    """

    __selector__ = selector(
        event_source="server", event_type="ubc.peer_instruction.original_submitted"
    )

    event: Union[
        Json[PeerInstructionEventField],
        PeerInstructionEventField,
    ]
    event_type: Literal["ubc.peer_instruction.original_submitted"]
    name: Literal["ubc.peer_instruction.original_submitted"]


class PeerInstructionRevisedSubmitted(BaseServerModel):
    """Pydantic model for `ubc.peer_instruction.revised_submitted` statement.

    The server emits this event when learners submit their revised responses. These
    events record the answer choice selected by the learner and the explanation for
    why that selection was made.

    Attributes:
        event (int): See PeerInstructionEventField.
        event_type (str): Consists of the value
            `ubc.peer_instruction.revised_submitted`.
        name (str): Consists of the value `ubc.peer_instruction.revised_submitted`.
    """

    __selector__ = selector(
        event_source="server", event_type="ubc.peer_instruction.revised_submitted"
    )

    event: Union[
        Json[PeerInstructionEventField],
        PeerInstructionEventField,
    ]
    event_type: Literal["ubc.peer_instruction.revised_submitted"]
    name: Literal["ubc.peer_instruction.revised_submitted"]



"""Virtual classroom xAPI events context fields definitions."""

import sys
from datetime import datetime
from typing import List, Optional, Union
from uuid import UUID

from pydantic import Field, validator

from ..base.contexts import BaseXapiContext, BaseXapiContextContextActivities
from ..base.unnested_objects import BaseXapiActivity
from ..concepts.activity_types.scorm_profile import ProfileActivity
from ..concepts.activity_types.virtual_classroom import VirtualClassroomActivity
from ..concepts.constants.cmi5_profile import CONTEXT_EXTENSION_SESSION_ID
from ..concepts.constants.tincan_vocabulary import CONTEXT_EXTENSION_PLANNED_DURATION
from ..config import BaseExtensionModelWithConfig

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class VirtualClassroomProfileActivity(ProfileActivity):
    """Pydantic model for virtual classroom profile `Activity` type.

    Attributes:
        id (str): Consists of the value `https://w3id.org/xapi/virtual-classroom`.
    """

    id: Literal[
        "https://w3id.org/xapi/virtual-classroom"
    ] = "https://w3id.org/xapi/virtual-classroom"


class VirtualClassroomContextContextActivities(BaseXapiContextContextActivities):
    """Pydantic model for virtual classroom `context`.`contextActivities` property.

    Attributes:
        category (dict or list): see VirtualClassroomProfileActivity.
    """

    category: Union[
        VirtualClassroomProfileActivity,
        List[Union[VirtualClassroomProfileActivity, BaseXapiActivity]],
    ]

    @validator("category")
    @classmethod
    def check_presence_of_profile_activity_category(
        cls,
        value: Union[
            VirtualClassroomProfileActivity,
            List[Union[VirtualClassroomProfileActivity, BaseXapiActivity]],
        ],
    ) -> Union[
        VirtualClassroomProfileActivity,
        List[Union[VirtualClassroomProfileActivity, BaseXapiActivity]],
    ]:
        """Check that the category list contains a `VirtualClassroomProfileActivity`."""
        if isinstance(value, VirtualClassroomProfileActivity):
            return value
        for activity in value:
            if isinstance(activity, VirtualClassroomProfileActivity):
                return value
        raise ValueError(
            "The `context.contextActivities.category` field should contain at least "
            "one valid `VirtualClassroomProfileActivity`"
        )


class VirtualClassroomContextExtensions(BaseExtensionModelWithConfig):
    """Pydantic model for virtual classroom base `context`.`extensions` property.

    Attributes:
        session_id (str): Consists of the ID of the active session.
    """

    session_id: str = Field(alias=CONTEXT_EXTENSION_SESSION_ID, default="")


class VirtualClassroomContext(BaseXapiContext):
    """Pydantic model for virtual classroom base `context` property.

    Attributes:
        registration (uuid): the registration that the Statement is associated with.
        extensions (dict): see VirtualClassroomContextExtensions.
    """

    contextActivities: VirtualClassroomContextContextActivities
    registration: UUID
    extensions: VirtualClassroomContextExtensions


# Initialized statement


class VirtualClassroomInitializedContextExtensions(VirtualClassroomContextExtensions):
    """Pydantic model for virtual classroom initialized `context`.`extensions` property.

    Attributes:
        planned_duration (datetime): Consists of the estimated duration of the scheduled
            virtual classroom.
    """

    planned_duration: Optional[datetime] = Field(
        alias=CONTEXT_EXTENSION_PLANNED_DURATION
    )


class VirtualClassroomInitializedContext(VirtualClassroomContext):
    """Pydantic model for virtual classroom initialized `context` property.

    Attributes:
        extensions (dict): see VirtualClassroomInitializedContextExtensions.
    """

    extensions: VirtualClassroomInitializedContextExtensions


# Joined statement


class VirtualClassroomJoinedContextExtensions(VirtualClassroomContextExtensions):
    """Pydantic model for virtual classroom initialized `context`.`extensions` property.

    Attributes:
        planned_duration (datetime): Consists of the estimated duration of the scheduled
            virtual classroom.
    """

    planned_duration: Optional[datetime] = Field(
        alias=CONTEXT_EXTENSION_PLANNED_DURATION
    )


class VirtualClassroomJoinedContext(VirtualClassroomContext):
    """Pydantic model for virtual classroom joined `context` property.

    Attributes:
        extensions (dict): see VirtualClassroomJoinedContextExtensions.
    """

    extensions: VirtualClassroomJoinedContextExtensions


# Terminated statement


class VirtualClassroomTerminatedContextExtensions(VirtualClassroomContextExtensions):
    """Pydantic model for virtual classroom terminated `context`.`extensions` property.

    Attributes:
        planned_duration (datetime): Consists of the estimated duration of the scheduled
            virtual classroom.
    """

    planned_duration: Optional[datetime] = Field(
        alias=CONTEXT_EXTENSION_PLANNED_DURATION
    )


class VirtualClassroomTerminatedContext(VirtualClassroomContext):
    """Pydantic model for virtual classroom terminated `context` property.

    Attributes:
        extensions (dict): see VirtualClassroomInitializedContextExtensions.
    """

    extensions: VirtualClassroomTerminatedContextExtensions


class VirtualClassroomStartedPollContextActivities(
    VirtualClassroomContextContextActivities
):
    """Pydantic model for virtual classroom started poll `context`.`contextActivities`
    property.

    Attributes:
        parent (list): see VirtualClassroomActivity.
    """  # noqa: D205

    parent: Union[VirtualClassroomActivity, List[VirtualClassroomActivity]]


class VirtualClassroomStartedPollContext(VirtualClassroomContext):
    """Pydantic model for virtual classroom started poll `context` property.

    Attributes:
        extensions (dict): see VirtualClassroomContextExtensions.
        contextActivities (dict): see
            VirtualClassroomStartedPollContextActivities.
    """

    extensions: VirtualClassroomContextExtensions
    contextActivities: VirtualClassroomStartedPollContextActivities


# Answered poll statement


class VirtualClassroomAnsweredPollContextActivities(
    VirtualClassroomContextContextActivities
):
    """Pydantic model for virtual classroom answered poll `context`.`contextActivities`
    property.

    Attributes:
        parent (list): see VirtualClassroomActivity.
    """  # noqa: D205

    parent: Union[VirtualClassroomActivity, List[VirtualClassroomActivity]]


class VirtualClassroomAnsweredPollContext(VirtualClassroomContext):
    """Pydantic model for virtual classroom answered poll `context` property.

    Attributes:
        extensions (dict): see VirtualClassroomContextExtensions.
        contextActivities (dict): see VirtualClassroomAnsweredPollContextActivities.
    """

    extensions: VirtualClassroomContextExtensions
    contextActivities: VirtualClassroomAnsweredPollContextActivities


# Posted public message statement


class VirtualClassroomPostedPublicMessageContextActivities(
    VirtualClassroomContextContextActivities
):
    """Pydantic model for virtual classroom posted public message
    `context`.`contextActivities` property.

    Attributes:
        parent (list): see VirtualClassroomActivity.
    """  # noqa: D205

    parent: Union[VirtualClassroomActivity, List[VirtualClassroomActivity]]


class VirtualClassroomPostedPublicMessageContext(VirtualClassroomContext):
    """Pydantic model for virtual classroom posted public message `context` property.

    Attributes:
        extensions (dict): see VirtualClassroomContextExtensions.
        contextActivities (list): see
            VirtualClassroomPostedPublicMessageContextActivities.
    """

    extensions: VirtualClassroomContextExtensions
    contextActivities: VirtualClassroomPostedPublicMessageContextActivities



"""Tests for the ralph.filters module."""

import pytest

from ralph import filters
from ralph.exceptions import EventKeyError


def test_filters_anonymous_with_empty_events():
    """Test the anonymous filter when input dict has not the expected `username`
    key.
    """
    event = {}
    with pytest.raises(EventKeyError):
        filters.anonymous(event)


def test_filters_anonymous_filtering():
    """Test anonymous filtering reliability."""
    event = {"username": "john"}
    anonymous_event = {"username": ""}
    assert filters.anonymous(event) == event
    assert filters.anonymous(anonymous_event) is None



"""Ralph CLI entrypoint."""

import json
import logging
import re
import sys
from inspect import isasyncgen, isclass
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any, Callable, Dict, Optional, Type, Union

import bcrypt

try:
    import click
except ModuleNotFoundError as err:
    raise ModuleNotFoundError(
        "You need to install 'cli' optional dependencies to use the ralph "
        "command: pip install ralph-malph[cli]"
    ) from err
try:
    import uvicorn
except ModuleNotFoundError:
    # This error will be caught in the runserver command. We should be able to
    # use all commands except the runserver command when lrs optional
    # dependencies are not installed.
    pass
from click_option_group import optgroup

from ralph import __version__ as ralph_version
from ralph.backends.data.base import (
    AsyncWritable,
    BaseAsyncDataBackend,
    BaseOperationType,
    BaseQuery,
)
from ralph.backends.loader import (
    get_cli_backends,
    get_cli_list_backends,
    get_cli_write_backends,
    get_lrs_backends,
)
from ralph.conf import ClientOptions, CommaSeparatedTuple, HeadersParameters, settings
from ralph.logger import configure_logging
from ralph.models.converter import Converter
from ralph.models.selector import ModelSelector
from ralph.models.validator import Validator
from ralph.utils import (
    execute_async,
    get_backend_class,
    get_backend_instance,
    get_root_logger,
    import_string,
    iter_over_async,
)

# cli module logger
logger = logging.getLogger(__name__)


class CommaSeparatedTupleParamType(click.ParamType):
    """Comma-separated tuple parameter type."""

    name = "value1,value2,value3"

    def convert(self, value, param, ctx):
        """Split the value by comma to return a tuple of values."""
        if isinstance(value, str):
            return tuple(value.split(","))

        if not isinstance(value, tuple):
            self.fail(
                "You should provide values separated by commas, e.g. foo,bar,baz",
                param,
                ctx,
            )

        return value


class CommaSeparatedKeyValueParamType(click.ParamType):
    """Comma-separated key=value parameter type."""

    name = "key=value,key=value"

    def convert(self, value, param, ctx):
        """Split the values by comma and equal sign.

        Return a dictionary build with key/value pairs.
        """
        if isinstance(value, dict):
            return value

        # Parse options string
        try:
            options = dict(option.split("=") for option in value.split(","))
        except ValueError:
            self.fail(
                (
                    "You should provide key=value pairs separated by commas, "
                    "e.g. foo=bar,bar=2"
                ),
                param,
                ctx,
            )

        # Cast simple types
        for key, value_ in options.items():
            if value_ == "":
                options.update({key: None})
            elif re.match("^true$", value_, re.IGNORECASE):
                options.update({key: True})
            elif re.match("^false$", value_, re.IGNORECASE):
                options.update({key: False})
            elif re.match(r"^[0-9]+\.[0-9]+$", value_):
                options.update({key: float(value_)})
            elif re.match("^[0-9]+$", value_):
                options.update({key: int(value_)})

        return options


class ClientOptionsParamType(CommaSeparatedKeyValueParamType):
    """Comma-separated key=value parameter type for client options."""

    def __init__(self, client_options_type: Any) -> None:
        """Instantiate ClientOptionsParamType for a client_options_type.

        Args:
            client_options_type (any): Pydantic model used for client options.
        """
        self.client_options_type = client_options_type

    def convert(self, value, param, ctx):
        """Split the values by comma and equal sign.

        Return an instance of client_options_type build with key/value pairs.
        """
        if isinstance(value, self.client_options_type):
            return value

        return self.client_options_type(**super().convert(value, param, ctx))


class HeadersParametersParamType(CommaSeparatedKeyValueParamType):
    """Comma-separated key=value parameter type for headers parameters."""

    def __init__(self, headers_parameters_type: Any) -> None:
        """Instantiate HeadersParametersParamType for a headers_parameters_type.

        Args:
            headers_parameters_type (any): Pydantic model used for headers parameters.
        """
        self.headers_parameters_type = headers_parameters_type

    def convert(self, value, param, ctx):
        """Split the values by comma and equal sign.

        Return an instance of headers_parameters_type build with key/value pairs.
        """
        if isinstance(value, self.headers_parameters_type):
            return value

        return self.headers_parameters_type(**super().convert(value, param, ctx))


class JSONStringParamType(click.ParamType):
    """JSON string parameter type."""

    name = '\'{"key": "value", "key": "value"}\''

    def convert(self, value, param, ctx):
        """Load value as a json string and return a dict."""
        try:
            options = json.loads(value)
        except (json.JSONDecodeError, TypeError):
            self.fail(
                "You should provide a valid JSON string as input",
                param,
                ctx,
            )
        return options


class RalphCLI(click.Group):
    """Ralph CLI entrypoint."""

    lazy_commands: Dict[str, Callable] = {}

    @classmethod
    def lazy_backends_options(
        cls, get_backends: Callable, name: Optional[str] = None
    ) -> Callable:
        """Lazy backend-related options decorator for Ralph commands."""

        def wrapper(command):
            command_name = name or command.__name__
            cls.lazy_commands[command_name] = lambda: backends_options(
                get_backends(), command_name
            )(command)
            return command

        return wrapper

    def invoke(self, ctx: click.Context):
        """Configure logging before click calls `list_commands` or `get_command`."""
        configure_logging()
        verbosity = ctx.params.get("verbosity")
        if verbosity is not None:
            level = getattr(logging, verbosity)
            get_root_logger().setLevel(level)
            for handler in get_root_logger().handlers:
                handler.setLevel(level)
        return super().invoke(ctx)

    def list_commands(self, ctx):
        """Register all lazy commands before calling `list_commands`."""
        for command in self.lazy_commands.values():
            command()
        self.lazy_commands = {}
        return super().list_commands(ctx)

    def get_command(self, ctx, cmd_name) -> Union[click.Command, None]:
        """Register lazy command (if it is requested) before calling `get_command`."""
        if cmd_name in self.lazy_commands:
            self.lazy_commands[cmd_name]()
            del self.lazy_commands[cmd_name]
        return super().get_command(ctx, cmd_name)


@click.group(name="ralph", cls=RalphCLI)
@click.option(
    "-v",
    "--verbosity",
    type=click.Choice(["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"]),
    metavar="LVL",
    required=False,
    help="Either CRITICAL, ERROR, WARNING, INFO (default) or DEBUG",
)
@click.version_option(version=ralph_version)
def cli(verbosity=None):  # noqa: ARG001
    """The cli is a stream-based tool to play with your logs.

    It offers functionalities to:
    - Validate or convert learning data in different standards
    - Read and write learning data to various databases or servers
    - Manage an instance of a Ralph LRS server
    """


# Once we have a base backend interface we could use Dict[str, Type[BaseBackend]]
def backends_options(backends: Dict[str, Type], name: Optional[str] = None):
    """Backend-related options decorator for Ralph commands."""

    def wrapper(command):
        backend_names = []
        for backend_name, backend in backends.items():
            backend_names.append(backend_name)
            fields = backend.settings_class.__fields__.items()
            for field_name, field in sorted(fields, key=lambda x: x[0], reverse=True):
                field_type = field.type_
                field_name = (  # noqa: PLW2901
                    f"{backend_name}-{field_name.lower()}".replace("_", "-")
                )
                option = f"--{field_name}"
                option_kwargs = {"default": None}
                if field.default:
                    option_kwargs["type"] = type(field.default)
                # If the field is a boolean, convert it to a flag option
                if field_type is bool:
                    option = f"{option}/--no-{field_name}"
                    option_kwargs["is_flag"] = True
                elif field_type is dict:
                    option_kwargs["type"] = CommaSeparatedKeyValueParamType()
                elif field_type is CommaSeparatedTuple:
                    option_kwargs["type"] = CommaSeparatedTupleParamType()
                elif isclass(field_type) and issubclass(field_type, ClientOptions):
                    option_kwargs["type"] = ClientOptionsParamType(field_type)
                elif isclass(field_type) and issubclass(field_type, HeadersParameters):
                    option_kwargs["type"] = HeadersParametersParamType(field_type)
                elif field_type is Path:
                    option_kwargs["type"] = click.Path()

                command = optgroup.option(option.lower(), **option_kwargs)(command)

            command = (optgroup.group(f"{backend_name} backend"))(command)

        command = click.option(
            "-b",
            "--backend",
            type=click.Choice(sorted(backend_names)),
            required=True,
            help="Backend",
        )(command)
        return cli.command(name=name or command.__name__)(command)

    return wrapper


@cli.command()
@click.option(
    "-u",
    "--username",
    type=str,
    required=True,
    help="The user for which we generate credentials.",
)
@click.password_option(
    "-p",
    "--password",
    type=str,
    required=True,
    help="The password to encrypt for this user. Will be prompted if missing.",
)
@click.option(
    "-s",
    "--scope",
    type=str,
    required=True,
    multiple=True,
    default=[],
    help="The user scope(s). This option can be provided multiple times.",
)
@click.option(
    "-t",
    "--target",
    type=str,
    required=False,
    help="The target location where statements are stored for the user.",
)
@click.option(
    "-M",
    "--agent-ifi-mbox",
    type=str,
    required=False,
    help="The mbox Inverse Functional Identifier of the associated agent.",
)
@click.option(
    "-S",
    "--agent-ifi-mbox-sha1sum",
    type=str,
    required=False,
    help="The mbox-sha1sum Inverse Functional Identifier of the associated agent.",
)
@click.option(
    "-O",
    "--agent-ifi-openid",
    type=str,
    required=False,
    help="The openid Inverse Functional Identifier of the associated agent.",
)
@click.option(
    "-A",
    "--agent-ifi-account",
    type=str,
    nargs=2,
    required=False,
    help=(
        'Input "{name} {homePage}". The account Inverse Functional Identifier of the '
        "associated agent."
    ),
)
@click.option(
    "-N",
    "--agent-name",
    type=str,
    required=False,
    help="The name of the associated agent.",
)
@click.option(
    "-w",
    "--write-to-disk",
    is_flag=True,
    default=False,
    help="Write new credentials to the LRS authentication file.",
)
def auth(  # noqa: PLR0913
    username,
    password,
    scope,
    target,
    write_to_disk,
    agent_ifi_mbox,
    agent_ifi_mbox_sha1sum,
    agent_ifi_openid,
    agent_ifi_account,
    agent_name,
):
    """Generate credentials for LRS HTTP basic authentication."""
    logger.info("Will generate credentials for user: %s", username)

    # Verify that exactly one agent representation has been provided
    if (
        sum(
            opt is not None
            for opt in [
                agent_ifi_mbox,
                agent_ifi_mbox_sha1sum,
                agent_ifi_openid,
                agent_ifi_account,
            ]
        )
        != 1
    ):
        raise click.UsageError(
            "Exactly one (1) of the ifi options is required. These options are:\n"
            "-M --agent-ifi-mbox : mbox Agent IFI\n"
            "-S --agent-ifi-mboxsha1sum : mbox-sha1sum Agent IFIr\n"
            "-O --agent-ifi-openid : openid Agent IFI\n"
            "-A --agent-ifi-account : account Agent IFI"
        )

    # Import required Pydantic models dynamically so that we don't create a
    # direct dependency between the CLI and the LRS

    ServerUsersCredentials = import_string(
        "ralph.api.auth.basic.ServerUsersCredentials"
    )
    UserCredentialsBasicAuth = import_string("ralph.api.auth.basic.UserCredentials")

    # NB: renaming classes below for clarity
    Account = import_string("ralph.models.xapi.base.ifi.BaseXapiAccount")
    AgentMbox = import_string("ralph.models.xapi.base.agents.BaseXapiAgentWithMbox")
    AgentMboxSha1sum = import_string(
        "ralph.models.xapi.base.agents.BaseXapiAgentWithMboxSha1Sum"
    )
    AgentOpenid = import_string("ralph.models.xapi.base.agents.BaseXapiAgentWithOpenId")
    AgentAccount = import_string(
        "ralph.models.xapi.base.agents.BaseXapiAgentWithAccount"
    )

    if agent_ifi_mbox:
        if agent_ifi_mbox[:7] != "mailto:":
            raise click.UsageError(
                'Mbox field must start with "mailto:" (e.g.: "mailto:foo@bar.com")'
            )
        agent = AgentMbox(mbox=agent_ifi_mbox, name=agent_name, objectType="Agent")
    if agent_ifi_mbox_sha1sum:
        agent = AgentMboxSha1sum(
            mbox_sha1sum=agent_ifi_mbox_sha1sum, name=agent_name, objectType="Agent"
        )
    if agent_ifi_openid:
        agent = AgentOpenid(
            openid=agent_ifi_openid, name=agent_name, objectType="Agent"
        )
    if agent_ifi_account:
        # Parse account details
        account = Account(homePage=agent_ifi_account[1], name=agent_ifi_account[0])
        agent = AgentAccount(account=account, name=agent_name, objectType="Agent")

    credentials = UserCredentialsBasicAuth(
        username=username,
        hash=bcrypt.hashpw(
            bytes(password, encoding=settings.LOCALE_ENCODING), bcrypt.gensalt()
        ).decode("ascii"),
        scopes=scope,
        target=target,
        agent=agent,
    )

    if write_to_disk:
        logger.info("Will append new credentials to: %s", settings.AUTH_FILE)

        # Force Path object instantiation so that the file creation can be
        # faked in a test environment.
        auth_file = Path(settings.AUTH_FILE)
        # Create the authentication file if it does not exist
        auth_file.parent.mkdir(parents=True, exist_ok=True)
        auth_file.touch()

        users = ServerUsersCredentials.parse_obj([])
        # Parse credentials file if not empty
        if auth_file.stat().st_size:
            users = ServerUsersCredentials.parse_file(auth_file)
        users += ServerUsersCredentials.parse_obj([credentials])
        auth_file.write_text(
            users.json(indent=2, exclude_none=True), encoding=settings.LOCALE_ENCODING
        )
        logger.info("User %s has been added to: %s", username, settings.AUTH_FILE)
    else:
        click.echo(
            (
                f"Copy/paste the following credentials to your LRS authentication "
                f"file located in: {settings.AUTH_FILE}\n"
                f"{credentials.json(indent=2, exclude_none=True)}"
            )
        )


@cli.command()
@click.option(
    "-p",
    "--parser",
    type=click.Choice([parser.lower() for parser, _ in settings.PARSERS]),
    required=True,
    help="Container format parser used to extract events",
)
def extract(parser):
    """Extract input events from a container format using a dedicated parser."""
    logger.info("Extracting events using the %s parser", parser)

    parser = getattr(settings.PARSERS, parser.upper()).get_instance()

    for event in parser.parse(sys.stdin):
        click.echo(event)


@cli.command()
@click.option(
    "-f",
    "--format",
    "format_",
    type=click.Choice(["edx", "xapi"]),
    required=True,
    help="Input events format to validate",
)
@click.option(
    "-I",
    "--ignore-errors",
    default=False,
    is_flag=True,
    help="Continue validating regardless of raised errors",
)
@click.option(
    "-F",
    "--fail-on-unknown",
    default=False,
    is_flag=True,
    help="Stop validating at first unknown event",
)
def validate(format_, ignore_errors, fail_on_unknown):
    """Validate input events of given format."""
    logger.info(
        "Validating %s events (ignore_errors=%s | fail-on-unknown=%s)",
        format_,
        ignore_errors,
        fail_on_unknown,
    )

    validator = Validator(ModelSelector(f"ralph.models.{format_}"))

    for event in validator.validate(sys.stdin, ignore_errors, fail_on_unknown):
        click.echo(event)


@cli.command()
@optgroup.group("From edX to xAPI converter options")
@optgroup.option(
    "-u",
    "--uuid-namespace",
    type=str,
    required=False,
    default=settings.CONVERTER_EDX_XAPI_UUID_NAMESPACE,
    help="The UUID namespace to use for the `ID` field generation",
)
@optgroup.option(
    "-p",
    "--platform-url",
    type=str,
    required=True,
    help="The `actor.account.homePage` to use in the xAPI statements",
)
@click.option(
    "-f",
    "--from",
    "from_",
    type=click.Choice(["edx"]),
    required=True,
    help="Input events format to convert",
)
@click.option(
    "-t",
    "--to",
    "to_",
    type=click.Choice(["xapi"]),
    required=True,
    help="Output events format",
)
@click.option(
    "-I",
    "--ignore-errors",
    default=False,
    is_flag=True,
    help="Continue writing regardless of raised errors",
)
@click.option(
    "-F",
    "--fail-on-unknown",
    default=False,
    is_flag=True,
    help="Stop converting at first unknown event",
)
def convert(from_, to_, ignore_errors, fail_on_unknown, **conversion_set_kwargs):
    """Convert input events to a given format."""
    logger.info(
        "Converting %s events to %s format (ignore_errors=%s | fail-on-unknown=%s)",
        from_,
        to_,
        ignore_errors,
        fail_on_unknown,
    )
    logger.debug("Converter parameters: %s", conversion_set_kwargs)

    converter = Converter(
        model_selector=ModelSelector(f"ralph.models.{from_}"),
        module=f"ralph.models.{from_}.converters.{to_}",
        **conversion_set_kwargs,
    )

    for event in converter.convert(sys.stdin, ignore_errors, fail_on_unknown):
        click.echo(event)


@RalphCLI.lazy_backends_options(get_cli_backends)
@click.argument("query", required=False)
@click.option(
    "-s",
    "--chunk-size",
    type=int,
    default=None,
    help="Get events by chunks of size #",
)
@click.option(
    "-t",
    "--target",
    type=str,
    default=None,
    help="Endpoint from which to read events (e.g. `/statements`)",
)
@click.option(
    "-i",
    "--ignore_errors",
    is_flag=False,
    show_default=True,
    default=False,
    help="Ignore errors during the encoding operation.",
)
def read(
    backend,
    chunk_size,
    target,
    query,
    ignore_errors,
    **options,
):
    """Read records matching the QUERY (json or string) from a configured backend."""
    logger.info(
        (
            "Fetching data from the configured %s backend "
            "(chunk size: %s | target: %s | query: %s)"
        ),
        backend,
        chunk_size,
        target,
        query,
    )
    logger.debug("Backend parameters: %s", options)

    backend_class = get_backend_class(get_cli_backends(), backend)
    backend = get_backend_instance(backend_class, options)

    if query and issubclass(backend.query_class, BaseQuery):
        query = backend.query_class.from_string(query)

    statements = backend.read(
        query=query,
        target=target,
        chunk_size=chunk_size,
        raw_output=True,
        ignore_errors=ignore_errors,
    )
    if isinstance(backend, BaseAsyncDataBackend):
        statements = iter_over_async(statements)

    for statement in statements:
        click.echo(statement, nl=False)


@RalphCLI.lazy_backends_options(get_cli_write_backends)
@click.option(
    "-t",
    "--target",
    type=str,
    default=None,
    help="The target container to write into",
)
@click.option(
    "-s",
    "--chunk-size",
    type=int,
    default=None,
    help="Get events by chunks of size #",
)
@click.option(
    "-I",
    "--ignore-errors",
    default=False,
    is_flag=True,
    help="Continue writing regardless of raised errors",
)
@click.option(
    "-o",
    "--operation-type",
    type=click.Choice([op_type.value for op_type in BaseOperationType]),
    metavar="OP_TYPE",
    required=False,
    help="Either index, create, delete, update or append",
)
@click.option(
    "-c",
    "--concurrency",
    default=1,
    help="Number of chunks to write concurrently. (async backends only)",
)
def write(  # noqa: PLR0913
    backend,
    target,
    chunk_size,
    ignore_errors,
    operation_type,
    concurrency,
    **options,
):
    """Write an archive to a configured backend."""
    logger.info("Writing to target %s for the configured %s backend", target, backend)

    logger.debug("Backend parameters: %s", options)

    backend_class = get_backend_class(get_cli_write_backends(), backend)
    backend = get_backend_instance(backend_class, options)

    writer = backend.write
    async_options = {}
    if isinstance(backend, AsyncWritable):
        writer = execute_async(backend.write)
        async_options = {"concurrency": concurrency}

    writer(
        data=sys.stdin.buffer,
        target=target,
        chunk_size=chunk_size,
        ignore_errors=ignore_errors,
        operation_type=BaseOperationType(operation_type) if operation_type else None,
        **async_options,
    )


@RalphCLI.lazy_backends_options(get_cli_list_backends, name="list")
@click.option(
    "-t",
    "--target",
    type=str,
    default=None,
    help="Container to list events from",
)
@click.option(
    "-n/-a",
    "--new/--all",
    default=False,
    help="List not fetched (or all) documents",
)
@click.option(
    "-D/-I",
    "--details/--ids",
    default=False,
    help="Get documents detailed output (JSON)",
)
def list_(target, details, new, backend, **options):
    """List available documents from a configured data backend."""
    logger.info("Listing documents for the configured %s backend", backend)
    logger.debug("Target container: %s", target)
    logger.debug("Fetch details: %s", str(details))
    logger.debug("Backend parameters: %s", options)

    backend_class = get_backend_class(get_cli_list_backends(), backend)
    backend = get_backend_instance(backend_class, options)

    documents = backend.list(target=target, details=details, new=new)
    documents = iter_over_async(documents) if isasyncgen(documents) else documents
    counter = 0
    for document in documents:
        click.echo(json.dumps(document) if details else document)
        counter += 1

    if counter == 0:
        logger.warning("Configured %s backend contains no document", backend.name)


@RalphCLI.lazy_backends_options(get_lrs_backends, name="runserver")
@click.option(
    "-h",
    "--host",
    type=str,
    required=False,
    default=settings.RUNSERVER_HOST,
    help="LRS server host name",
)
@click.option(
    "-p",
    "--port",
    type=int,
    required=False,
    default=settings.RUNSERVER_PORT,
    help="LRS server port",
)
def runserver(backend: str, host: str, port: int, **options):
    """Run the API server for the development environment.

    Starts uvicorn programmatically for convenience and documentation.
    """
    logger.info("Running API server on %s:%s with %s backend", host, port, backend)
    logger.info(
        (
            "Do not use runserver in production - start production servers "
            "through a process manager such as gunicorn/supervisor/circus."
        )
    )

    # The LRS server relies only on environment and configuration variables for its
    # configuration (not CLI arguments). Therefore, we convert the CLI arguments to
    # environment variables by creating a temporary environment file and passing it to
    # uvicorn.
    with NamedTemporaryFile(mode="w", encoding=settings.LOCALE_ENCODING) as env_file:
        env_file.write(f"RALPH_RUNSERVER_BACKEND={backend}\n")
        for key, value in options.items():
            if value is None:
                continue
            backend_name, field_name = key.split(sep="_", maxsplit=1)
            key = (  # noqa: PLW2901
                f"RALPH_BACKENDS__LRS__{backend_name}__{field_name}".upper()
            )
            if isinstance(value, tuple):
                value = ",".join(value)  # noqa: PLW2901
            if issubclass(type(value), ClientOptions):
                for key_dict, value_dict in value.dict().items():
                    if value_dict is None:
                        continue
                    key_dict = f"{key}__{key_dict}"  # noqa: PLW2901
                    logger.debug(
                        "Setting environment variable %s to '%s'", key_dict, value_dict
                    )
                    env_file.write(f"{key_dict}={value_dict}\n")
                continue
            logger.debug("Setting environment variable %s to '%s'", key, value)
            env_file.write(f"{key}={value}\n")
        env_file.seek(0)
        try:
            uvicorn.run(
                "ralph.api:app",
                env_file=env_file.name,
                host=host,
                port=port,
                log_config=settings.LOGGING,
                log_level="debug",
                reload=True,
            )
        except NameError as error:
            raise ModuleNotFoundError(
                "You need to install 'lrs' optional dependencies to use the runserver "
                "command: pip install ralph-malph[lrs]"
            ) from error

    logger.info("Shutting down uvicorn server.")



"""Asynchronous Elasticsearch data backend for Ralph."""

import logging
from io import IOBase
from typing import AsyncIterator, Iterable, Optional, TypeVar, Union

from elasticsearch import ApiError, AsyncElasticsearch, TransportError
from elasticsearch.helpers import BulkIndexError, async_streaming_bulk
from pydantic import PositiveInt

from ralph.backends.data.base import (
    AsyncListable,
    AsyncWritable,
    BaseAsyncDataBackend,
    BaseOperationType,
    DataBackendStatus,
)
from ralph.backends.data.es import ESDataBackend, ESDataBackendSettings, ESQuery
from ralph.exceptions import BackendException

logger = logging.getLogger(__name__)
Settings = TypeVar("Settings", bound=ESDataBackendSettings)


class AsyncESDataBackend(
    BaseAsyncDataBackend[Settings, ESQuery], AsyncWritable, AsyncListable
):
    """Asynchronous Elasticsearch data backend."""

    name = "async_es"
    unsupported_operation_types = {BaseOperationType.APPEND}

    def __init__(self, settings: Optional[Settings] = None):
        """Instantiate the asynchronous Elasticsearch client.

        Args:
            settings (ESDataBackendSettings or None): The data backend settings.
                If `settings` is `None`, a default settings instance is used instead.
        """
        super().__init__(settings)
        self._client = None

    @property
    def client(self) -> AsyncElasticsearch:
        """Create an AsyncElasticsearch client if it doesn't exist."""
        if not self._client:
            self._client = AsyncElasticsearch(
                self.settings.HOSTS, **self.settings.CLIENT_OPTIONS.dict()
            )
        return self._client

    async def status(self) -> DataBackendStatus:
        """Check Elasticsearch cluster connection and status."""
        try:
            await self.client.info()
            cluster_status = await self.client.cat.health()
        except TransportError as error:
            logger.error("Failed to connect to Elasticsearch: %s", error)
            return DataBackendStatus.AWAY

        if "green" in cluster_status:
            return DataBackendStatus.OK

        if "yellow" in cluster_status and self.settings.ALLOW_YELLOW_STATUS:
            logger.info("Cluster status is yellow.")
            return DataBackendStatus.OK

        logger.error("Cluster status is not green: %s", cluster_status)

        return DataBackendStatus.ERROR

    async def list(
        self, target: Optional[str] = None, details: bool = False, new: bool = False
    ) -> Union[AsyncIterator[str], AsyncIterator[dict]]:
        """List available Elasticsearch indices, data streams and aliases.

        Args:
            target (str or None): The comma-separated list of data streams, indices,
                and aliases to limit the request. Supports wildcards (*).
                If target is `None`, lists all available indices, data streams and
                    aliases. Equivalent to (`target` = "*").
            details (bool): Get detailed information instead of just names.
            new (bool): Ignored.

        Yield:
            str: The next index, data stream or alias name. (If `details` is False).
            dict: The next index, data stream or alias details. (If `details` is True).

        Raise:
            BackendException: If a failure during indices retrieval occurs.
        """
        target = target if target else "*"
        try:
            indices = await self.client.indices.get(index=target)
        except (ApiError, TransportError) as error:
            msg = "Failed to read indices: %s"
            logger.error(msg, error)
            raise BackendException(msg % error) from error

        if new:
            logger.warning("The `new` argument is ignored")

        if details:
            for index, value in indices.items():
                yield {index: value}

            return

        for index in indices:
            yield index

    async def read(  # noqa: PLR0913
        self,
        query: Optional[ESQuery] = None,
        target: Optional[str] = None,
        chunk_size: Optional[int] = None,
        raw_output: bool = False,
        ignore_errors: bool = False,
        prefetch: Optional[PositiveInt] = None,
        max_statements: Optional[PositiveInt] = None,
    ) -> Union[AsyncIterator[bytes], AsyncIterator[dict]]:
        """Read documents matching the query in the target index and yield them.

        Args:
            query (ESQuery): The Elasticsearch query to use when fetching documents.
            target (str or None): The target Elasticsearch index name to query.
                If target is `None`, the `DEFAULT_INDEX` is used instead.
            chunk_size (int or None): The chunk size when reading documents by batches.
                If `chunk_size` is `None` it defaults to `READ_CHUNK_SIZE`.
            raw_output (bool): Controls whether to yield dictionaries or bytes.
            ignore_errors (bool): No impact as encoding errors are not expected in
                Elasticsearch results.
            prefetch (int): The number of records to prefetch (queue) while yielding.
                If `prefetch` is `None` it defaults to `1`, i.e. no records are
                prefetched.
            max_statements (int): The maximum number of statements to yield.
                If `None` (default) or `0`, there is no maximum.

        Yield:
            bytes: The next raw document if `raw_output` is True.
            dict: The next JSON parsed document if `raw_output` is False.

        Raise:
            BackendException: If a failure occurs during Elasticsearch connection.
        """
        statements = super().read(
            query,
            target,
            chunk_size,
            raw_output,
            ignore_errors,
            prefetch,
            max_statements,
        )
        async for statement in statements:
            yield statement

    async def _read_dicts(
        self,
        query: ESQuery,
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,  # noqa: ARG002
    ) -> AsyncIterator[dict]:
        """Method called by `self.read` yielding dictionaries. See `self.read`."""
        target = target if target else self.settings.DEFAULT_INDEX
        if not query.pit.keep_alive:
            query.pit.keep_alive = self.settings.POINT_IN_TIME_KEEP_ALIVE
        if not query.pit.id:
            try:
                query.pit.id = (
                    await self.client.open_point_in_time(
                        index=target, keep_alive=query.pit.keep_alive
                    )
                )["id"]
            except (ApiError, TransportError, ValueError) as error:
                msg = "Failed to open Elasticsearch point in time: %s"
                logger.error(msg, error)
                raise BackendException(msg % error) from error

        limit = query.size
        kwargs = query.dict()
        count = chunk_size
        # The first condition is set to comprise either limit as None
        # (when the backend query does not have `size` parameter),
        # or limit with a positive value.
        while limit != 0 and chunk_size == count:
            kwargs["size"] = limit if limit and limit < chunk_size else chunk_size
            try:
                documents = (await self.client.search(**kwargs))["hits"]["hits"]
            except (ApiError, TransportError, TypeError) as error:
                msg = "Failed to execute Elasticsearch query: %s"
                logger.error(msg, error)
                raise BackendException(msg % error) from error
            count = len(documents)
            if limit:
                limit -= count if chunk_size == count else limit
            query.search_after = None
            if count:
                query.search_after = [str(part) for part in documents[-1]["sort"]]
            kwargs["search_after"] = query.search_after
            for document in documents:
                yield document

    async def write(  # noqa: PLR0913
        self,
        data: Union[IOBase, Iterable[bytes], Iterable[dict]],
        target: Optional[str] = None,
        chunk_size: Optional[int] = None,
        ignore_errors: bool = False,
        operation_type: Optional[BaseOperationType] = None,
        concurrency: Optional[PositiveInt] = None,
    ) -> int:
        """Write data documents to the target index and return their count.

        Args:
            data (Iterable or IOBase): The data containing documents to write.
            target (str or None): The target Elasticsearch index name.
                If target is `None`, the `DEFAULT_INDEX` is used instead.
            chunk_size (int or None): The number of documents to write in one batch.
                If `chunk_size` is `None` it defaults to `WRITE_CHUNK_SIZE`.
            ignore_errors (bool): If `True`, errors during decoding, encoding and
                sending batches of documents are ignored and logged.
                If `False` (default), a `BackendException` is raised on any error.
            operation_type (BaseOperationType or None): The mode of the write operation.
                If `operation_type` is `None`, the `default_operation_type` is used
                instead. See `BaseOperationType`.
            concurrency (int): The number of chunks to write concurrently.
                If `None` it defaults to `1`.

        Return:
            int: The number of documents written.

        Raise:
            BackendException: If any failure occurs during the write operation or
                if an inescapable failure occurs and `ignore_errors` is set to `True`.
            BackendParameterException: If the `operation_type` is `APPEND` as it is not
                supported.
        """
        return await super().write(
            data, target, chunk_size, ignore_errors, operation_type, concurrency
        )

    async def _write_dicts(  # noqa: PLR0913
        self,
        data: Iterable[dict],
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,
        operation_type: BaseOperationType,
    ) -> int:
        """Method called by `self.write` writing dictionaries. See `self.write`."""
        count = 0
        target = target if target else self.settings.DEFAULT_INDEX
        msg = "Start writing to the %s index (chunk size: %d)"
        logger.debug(msg, target, chunk_size)
        try:
            async for success, action in async_streaming_bulk(
                client=self.client,
                actions=ESDataBackend.to_documents(data, target, operation_type),
                chunk_size=chunk_size,
                raise_on_error=(not ignore_errors),
                refresh=self.settings.REFRESH_AFTER_WRITE,
            ):
                count += success
                logger.debug("Wrote %d document [action: %s]", success, action)

            logger.info("Finished writing %d documents with success", count)
        except (BulkIndexError, ApiError, TransportError) as error:
            msg = "%s %s Total succeeded writes: %s"
            details = getattr(error, "errors", "")
            logger.error(msg, error, details, count)
            raise BackendException(msg % (error, details, count)) from error
        return count

    async def close(self) -> None:
        """Close the AsyncElasticsearch client.

        Raise:
            BackendException: If a failure occurs during the close operation.
        """
        if not self._client:
            logger.warning("No backend client to close.")
            return

        try:
            await self.client.close()
        except TransportError as error:
            msg = "Failed to close Elasticsearch client: %s"
            logger.error(msg, error)
            raise BackendException(msg % error) from error



"""edX pydantic models."""

# flake8: noqa

from ralph.models.edx.navigational import statements

from .enrollment.statements import (
    EdxCourseEnrollmentActivated,
    EdxCourseEnrollmentDeactivated,
    EdxCourseEnrollmentModeChanged,
    EdxCourseEnrollmentUpgradeSucceeded,
    UIEdxCourseEnrollmentUpgradeClicked,
)
from .navigational.statements import UIPageClose, UISeqGoto, UISeqNext, UISeqPrev
from .open_response_assessment.statements import (
    ORACreateSubmission,
    ORAGetPeerSubmission,
    ORAGetSubmissionForStaffGrading,
    ORAPeerAssess,
    ORASaveSubmission,
    ORASelfAssess,
    ORAStaffAssess,
    ORAStudentTrainingAssessExample,
    ORASubmitFeedbackOnAssessments,
    ORAUploadFile,
)
from .peer_instruction.statements import (
    PeerInstructionAccessed,
    PeerInstructionOriginalSubmitted,
    PeerInstructionRevisedSubmitted,
)
from .problem_interaction.statements import (
    EdxProblemHintDemandhintDisplayed,
    EdxProblemHintFeedbackDisplayed,
    ProblemCheck,
    ProblemCheckFail,
    ProblemRescore,
    ProblemRescoreFail,
    ResetProblem,
    ResetProblemFail,
    SaveProblemFail,
    SaveProblemSuccess,
    ShowAnswer,
    UIProblemCheck,
    UIProblemGraded,
    UIProblemReset,
    UIProblemSave,
    UIProblemShow,
)
from .server import Server
from .textbook_interaction.statements import (
    UIBook,
    UITextbookPdfChapterNavigated,
    UITextbookPdfDisplayScaled,
    UITextbookPdfOutlineToggled,
    UITextbookPdfPageNavigated,
    UITextbookPdfPageScrolled,
    UITextbookPdfSearchCaseSensitivityToggled,
    UITextbookPdfSearchExecuted,
    UITextbookPdfSearchHighlightToggled,
    UITextbookPdfSearchNavigatedNext,
    UITextbookPdfThumbnailNavigated,
    UITextbookPdfThumbnailsToggled,
    UITextbookPdfZoomButtonsChanged,
    UITextbookPdfZoomMenuChanged,
)
from .video.statements import (
    UIHideTranscript,
    UILoadVideo,
    UIPauseVideo,
    UIPlayVideo,
    UISeekVideo,
    UIShowTranscript,
    UISpeedChangeVideo,
    UIStopVideo,
    UIVideoHideCCMenu,
    UIVideoShowCCMenu,
)



"""`ADL Vocabulary` verbs definitions."""

import sys
from typing import Dict, Optional

from ...base.verbs import BaseXapiVerb
from ...constants import LANG_EN_US_DISPLAY

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class AskedVerb(BaseXapiVerb):
    """Pydantic model for asked `verb`.

    Attributes:
        id (str): Consists of the value `http://adlnet.gov/expapi/verbs/asked`.
        display (dict): Consists of the dictionary `{"en-US": "asked"}`.
    """

    id: Literal[
        "http://adlnet.gov/expapi/verbs/asked"
    ] = "http://adlnet.gov/expapi/verbs/asked"
    display: Optional[Dict[Literal[LANG_EN_US_DISPLAY], Literal["asked"]]]


class AnsweredVerb(BaseXapiVerb):
    """Pydantic model for answered `verb`.

    Attributes:
        id (str): Consists of the value `http://adlnet.gov/expapi/verbs/answered`.
        display (dict): Consists of the dictionary `{"en-US": "answered"}`.
    """

    id: Literal[
        "http://adlnet.gov/expapi/verbs/answered"
    ] = "http://adlnet.gov/expapi/verbs/answered"
    display: Optional[Dict[Literal[LANG_EN_US_DISPLAY], Literal["answered"]]]


class RegisteredVerb(BaseXapiVerb):
    """Pydantic model for registered `verb`.

    Attributes:
        id (str): Consists of the value `http://adlnet.gov/expapi/verbs/registered`.
        display (dict): Consists of the dictionary `{"en-US": "registered"}`.
    """

    id: Literal[
        "http://adlnet.gov/expapi/verbs/registered"
    ] = "http://adlnet.gov/expapi/verbs/registered"
    display: Optional[Dict[Literal[LANG_EN_US_DISPLAY], Literal["registered"]]]



"""Navigational event xAPI Converter."""
from typing import Set

from ralph.models.converter import ConversionItem
from ralph.models.edx.navigational.statements import UIPageClose
from ralph.models.xapi.navigation.statements import PageTerminated

from .base import BaseXapiConverter


class UIPageCloseToPageTerminated(BaseXapiConverter):
    """Convert a common edX `page_close` event to xAPI.

    Example Statement: John terminated https://www.fun-mooc.fr/ page.

    WARNING: The converter does not use the `self.platform_url` in the `object__id`
    because the `platform_url` is present in the edX's event `page` field.
    """

    __src__ = UIPageClose
    __dest__ = PageTerminated

    def _get_conversion_items(self) -> Set[ConversionItem]:
        """Return a set of ConversionItems used for conversion."""
        conversion_items = super()._get_conversion_items()
        return conversion_items.union({ConversionItem("object__id", "page")})



"""Base xAPI `Attachments` definitions."""

from typing import Optional

from pydantic import AnyUrl

from ..config import BaseModelWithConfig
from .common import IRI, LanguageMap


class BaseXapiAttachment(BaseModelWithConfig):
    """Pydantic model for `attachment` property.

    Attributes:
        usageType (IRI): Identifies the usage of this Attachment.
        display (LanguageMap): Consists of the Attachment's title.
        description (LanguageMap): Consists of the Attachment's description.
        contentType (str): Consists of the Attachment's content type.
        length (int): Consists of the length of the Attachment's data in octets.
        sha2 (str): Consists of the SHA-2 hash of the Attachment data.
        fileUrl (URL): Consists of the URL from which the Attachment can be retrieved.
    """

    usageType: IRI
    display: LanguageMap
    description: Optional[LanguageMap]
    contentType: str
    length: int
    sha2: str
    fileUrl: Optional[AnyUrl]



"""`Activity streams vocabulary` activity types definitions."""

import sys

from ...base.unnested_objects import BaseXapiActivity, BaseXapiActivityDefinition

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


# Page


class PageActivityDefinition(BaseXapiActivityDefinition):
    """Pydantic model for page `Activity` type `definition` property.

    Attributes:
       type (str): Consists of the value `http://activitystrea.ms/schema/1.0/page`.
    """

    type: Literal[
        "http://activitystrea.ms/schema/1.0/page"
    ] = "http://activitystrea.ms/schema/1.0/page"


class PageActivity(BaseXapiActivity):
    """Pydantic model for page `Activity` type.

    Attributes:
        definition (dict): See PageActivityDefinition.
    """

    definition: PageActivityDefinition = PageActivityDefinition()


# File


class FileActivityDefinition(BaseXapiActivityDefinition):
    """Pydantic model for file `Activity` type `definition` property.

    Attributes:
       type (str): Consists of the value `http://activitystrea.ms/file`.
    """

    type: Literal["http://activitystrea.ms/file"]


class FileActivity(BaseXapiActivity):
    """Pydantic model for file `Activity` type.

    Attributes:
        definition (dict): See FileActivityDefinition.
    """

    definition: FileActivityDefinition



"""Video event xAPI Converter."""
from typing import Set

from ralph.models.converter import ConversionItem
from ralph.models.edx.video.statements import (
    UILoadVideo,
    UIPauseVideo,
    UIPlayVideo,
    UISeekVideo,
    UIStopVideo,
)
from ralph.models.xapi.concepts.constants.video import (
    CONTEXT_EXTENSION_LENGTH,
    CONTEXT_EXTENSION_SESSION_ID,
    CONTEXT_EXTENSION_USER_AGENT,
    RESULT_EXTENSION_PROGRESS,
    RESULT_EXTENSION_TIME,
    RESULT_EXTENSION_TIME_FROM,
    RESULT_EXTENSION_TIME_TO,
)
from ralph.models.xapi.constants import LANG_EN_US_DISPLAY
from ralph.models.xapi.video.statements import (
    VideoInitialized,
    VideoPaused,
    VideoPlayed,
    VideoSeeked,
    VideoTerminated,
)

from .base import BaseXapiConverter


class VideoBaseXapiConverter(BaseXapiConverter):
    """Base Video xAPI Converter."""

    def _get_conversion_items(self) -> Set[ConversionItem]:
        """Return a set of ConversionItems used for conversion."""
        conversion_items = super()._get_conversion_items()
        return conversion_items.union(
            {
                ConversionItem(
                    "object__definition__name",
                    "event__id",
                    lambda id: {LANG_EN_US_DISPLAY: id},
                ),
                ConversionItem(
                    "object__id",
                    None,
                    lambda event: self.platform_url
                    + "/xblock/block-v1:"
                    + event["context"]["course_id"]
                    + "-course-v1:+type@video+block@"
                    + event["event"]["id"],
                ),
                ConversionItem(
                    "context__contextActivities__category",
                    None,
                    lambda _: [{"id": "https://w3id.org/xapi/video"}],
                ),
                ConversionItem(
                    "context__extensions__" + CONTEXT_EXTENSION_SESSION_ID,
                    "session",
                ),
            },
        )


class UILoadVideoToVideoInitialized(VideoBaseXapiConverter):
    """Convert a common edX `load_video` event to xAPI."""

    __src__ = UILoadVideo
    __dest__ = VideoInitialized

    def _get_conversion_items(self) -> Set[ConversionItem]:
        """Return a set of ConversionItems used for conversion."""
        conversion_items = super()._get_conversion_items()
        return conversion_items.union(
            {
                ConversionItem(
                    "context__extensions__" + CONTEXT_EXTENSION_LENGTH,
                    None,
                    # Set the video length to null by default.
                    # This information is mandatory in the xAPI template
                    # and does not exist in the edX `load_video` event model.
                    lambda _: 0.0,
                ),
                ConversionItem(
                    "context__extensions__" + CONTEXT_EXTENSION_SESSION_ID,
                    "session",
                ),
                ConversionItem(
                    "context__extensions__" + CONTEXT_EXTENSION_USER_AGENT, "agent"
                ),
            },
        )


class UIPlayVideoToVideoPlayed(VideoBaseXapiConverter):
    """Convert a common edX `play_video` event to xAPI."""

    __src__ = UIPlayVideo
    __dest__ = VideoPlayed

    def _get_conversion_items(self) -> Set[ConversionItem]:
        """Return a set of ConversionItems used for conversion."""
        conversion_items = super()._get_conversion_items()
        return conversion_items.union(
            {
                ConversionItem(
                    "result__extensions__" + RESULT_EXTENSION_TIME,
                    "event__currentTime",
                ),
                ConversionItem(
                    "context__extensions__" + CONTEXT_EXTENSION_SESSION_ID,
                    "session",
                ),
            },
        )


class UIPauseVideoToVideoPaused(VideoBaseXapiConverter):
    """Convert a common edX `pause_video` event to xAPI."""

    __src__ = UIPauseVideo
    __dest__ = VideoPaused

    def _get_conversion_items(self) -> Set[ConversionItem]:
        """Return a set of ConversionItems used for conversion."""
        conversion_items = super()._get_conversion_items()
        return conversion_items.union(
            {
                ConversionItem(
                    "result__extensions__" + RESULT_EXTENSION_TIME,
                    "event__currentTime",
                ),
                ConversionItem(
                    "context__extensions__" + CONTEXT_EXTENSION_LENGTH,
                    None,
                    # Set the video length to null by default.
                    # This information is mandatory in the xAPI template
                    # and does not exist in the edX `pause_video` event model.
                    lambda _: 0.0,
                ),
                ConversionItem(
                    "context__extensions__" + CONTEXT_EXTENSION_SESSION_ID,
                    "session",
                ),
            },
        )


class UIStopVideoToVideoTerminated(VideoBaseXapiConverter):
    """Convert a common edX `stop_video` event to xAPI."""

    __src__ = UIStopVideo
    __dest__ = VideoTerminated

    def _get_conversion_items(self) -> Set[ConversionItem]:
        """Return a set of ConversionItems used for conversion."""
        conversion_items = super()._get_conversion_items()
        return conversion_items.union(
            {
                ConversionItem(
                    "result__extensions__" + RESULT_EXTENSION_TIME,
                    "event__currentTime",
                ),
                ConversionItem(
                    "result__extensions__" + RESULT_EXTENSION_PROGRESS,
                    None,
                    # Set the video progress to null by default.
                    # This information is mandatory in the xAPI template
                    # and does not exist in the edX `stop_video` event model.
                    lambda _: 0.0,
                ),
                ConversionItem(
                    "context__extensions__" + CONTEXT_EXTENSION_LENGTH,
                    None,
                    # Set the video length to null by default.
                    # This information is mandatory in the xAPI template
                    # and does not exist in the edX `stop_video` event model.
                    lambda _: 0.0,
                ),
                ConversionItem(
                    "context__extensions__" + CONTEXT_EXTENSION_SESSION_ID,
                    "session",
                ),
            },
        )


class UISeekVideoToVideoSeeked(VideoBaseXapiConverter):
    """Convert a common edX `seek_video` event to xAPI."""

    __src__ = UISeekVideo
    __dest__ = VideoSeeked

    def _get_conversion_items(self) -> Set[ConversionItem]:
        """Return a set of ConversionItems used for conversion."""
        conversion_items = super()._get_conversion_items()
        return conversion_items.union(
            {
                ConversionItem(
                    "result__extensions__" + RESULT_EXTENSION_TIME_FROM,
                    "event__old_time",
                ),
                ConversionItem(
                    "result__extensions__" + RESULT_EXTENSION_TIME_TO,
                    "event__new_time",
                ),
                ConversionItem(
                    "context__extensions__" + CONTEXT_EXTENSION_SESSION_ID,
                    "session",
                ),
            },
        )



"""`Virtual classroom` verbs definitions."""

import sys
from typing import Dict, Optional

from ...base.verbs import BaseXapiVerb
from ...constants import LANG_EN_US_DISPLAY

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class MutedVerb(BaseXapiVerb):
    """Pydantic model for muted `verb`.

    Attributes:
        id (str): Consists of the value
            `https://w3id.org/xapi/virtual-classroom/verbs/muted`.
        display (dict): Consists of the dictionary `{"en-US": "muted"}`.
    """

    id: Literal[
        "https://w3id.org/xapi/virtual-classroom/verbs/muted"
    ] = "https://w3id.org/xapi/virtual-classroom/verbs/muted"
    display: Optional[Dict[Literal[LANG_EN_US_DISPLAY], Literal["muted"]]]


class UnmutedVerb(BaseXapiVerb):
    """Pydantic model for unmuted `verb`.

    Attributes:
        id (str): Consists of the value
            `https://w3id.org/xapi/virtual-classroom/verbs/unmuted`.
        display (dict): Consists of the dictionary `{"en-US": "unmuted"}`.
    """

    id: Literal[
        "https://w3id.org/xapi/virtual-classroom/verbs/unmuted"
    ] = "https://w3id.org/xapi/virtual-classroom/verbs/unmuted"
    display: Optional[Dict[Literal[LANG_EN_US_DISPLAY], Literal["unmuted"]]]


class StartedCameraVerb(BaseXapiVerb):
    """Pydantic model for started camera `verb`.

    Attributes:
        id (str): Consists of the value
            `https://w3id.org/xapi/virtual-classroom/verbs/started-camera`.
        display (dict): Consists of the dictionary `{"en-US": "started camera"}`.
    """

    id: Literal[
        "https://w3id.org/xapi/virtual-classroom/verbs/started-camera"
    ] = "https://w3id.org/xapi/virtual-classroom/verbs/started-camera"
    display: Optional[Dict[Literal[LANG_EN_US_DISPLAY], Literal["started camera"]]]


class StoppedCameraVerb(BaseXapiVerb):
    """Pydantic model for stopped camera `verb`.

    Attributes:
        id (str): Consists of the value
            `https://w3id.org/xapi/virtual-classroom/verbs/stopped-camera`.
        display (dict): Consists of the dictionary `{"en-US": "stopped camera"}`.
    """

    id: Literal[
        "https://w3id.org/xapi/virtual-classroom/verbs/stopped-camera"
    ] = "https://w3id.org/xapi/virtual-classroom/verbs/stopped-camera"
    display: Optional[Dict[Literal[LANG_EN_US_DISPLAY], Literal["stopped camera"]]]


class SharedScreenVerb(BaseXapiVerb):
    """Pydantic model for shared screen `verb`.

    Attributes:
        id (str): Consists of the value
            `https://w3id.org/xapi/virtual-classroom/verbs/shared-screen`.
        display (dict): Consists of the dictionary `{"en-US": "shared screen"}`.
    """

    id: Literal[
        "https://w3id.org/xapi/virtual-classroom/verbs/shared-screen"
    ] = "https://w3id.org/xapi/virtual-classroom/verbs/shared-screen"
    display: Optional[Dict[Literal[LANG_EN_US_DISPLAY], Literal["shared screen"]]]


class UnsharedScreenVerb(BaseXapiVerb):
    """Pydantic model for unshared screen `verb`.

    Attributes:
        id (str): Consists of the value
            `https://w3id.org/xapi/virtual-classroom/verbs/unshared-screen`.
        display (dict): Consists of the dictionary `{"en-US": "unshared screen"}`.
    """

    id: Literal[
        "https://w3id.org/xapi/virtual-classroom/verbs/unshared-screen"
    ] = "https://w3id.org/xapi/virtual-classroom/verbs/unshared-screen"
    display: Optional[Dict[Literal[LANG_EN_US_DISPLAY], Literal["unshared screen"]]]


class RaisedHandVerb(BaseXapiVerb):
    """Pydantic model for raised hand `verb`.

    Attributes:
        id (str): Consists of the value
            `https://w3id.org/xapi/virtual-classroom/verbs/raised-hand`.
        display (dict): Consists of the dictionary `{"en-US": "raised hand"}`.
    """

    id: Literal[
        "https://w3id.org/xapi/virtual-classroom/verbs/raised-hand"
    ] = "https://w3id.org/xapi/virtual-classroom/verbs/raised-hand"
    display: Optional[Dict[Literal[LANG_EN_US_DISPLAY], Literal["raised hand"]]]


class LoweredHandVerb(BaseXapiVerb):
    """Pydantic model for lowered hand `verb`.

    Attributes:
        id (str): Consists of the value
            `https://w3id.org/xapi/virtual-classroom/verbs/lowered-hand`.
        display (dict): Consists of the dictionary `{"en-US": "lowered hand"}`.
    """

    id: Literal[
        "https://w3id.org/xapi/virtual-classroom/verbs/lowered-hand"
    ] = "https://w3id.org/xapi/virtual-classroom/verbs/lowered-hand"
    display: Optional[Dict[Literal[LANG_EN_US_DISPLAY], Literal["lowered hand"]]]



"""Tests for Ralph utils."""

import logging

import pytest
from pydantic import BaseModel

from ralph import utils as ralph_utils
from ralph.exceptions import UnsupportedBackendException


def test_utils_import_string():
    """Test import_string utility taken from Django utilities."""
    with pytest.raises(ImportError, match="foo doesn't look like a module path"):
        ralph_utils.import_string("foo")

    with pytest.raises(
        ImportError, match='Module "requests" does not define a "foo" attribute/class'
    ):
        ralph_utils.import_string("requests.foo")

    http_status = ralph_utils.import_string("http.HTTPStatus")
    assert http_status.OK == 200


def test_utils_get_backend_class(caplog):
    """Test the `get_backend_class` utility should return the expected result."""
    backends = {"es": type("es_backend", (), {}), "fs": type("fs_backend", (), {})}
    assert ralph_utils.get_backend_class(backends, "es") == backends["es"]
    assert ralph_utils.get_backend_class(backends, "fs") == backends["fs"]
    msg = "'foo' backend not found in available backends: es, fs"
    with caplog.at_level(logging.ERROR):
        with pytest.raises(UnsupportedBackendException, match=msg):
            ralph_utils.get_backend_class(backends, "foo")

    assert ("ralph.utils", logging.ERROR, msg) in caplog.record_tuples


@pytest.mark.parametrize(
    "options,expected",
    [
        # Empty options should produce default result.
        ({}, {"FOO": "FOO"}),
        # Options not matching the backend name are ignored.
        ({"foo": "bar", "not_dummy_foo": "baz"}, {"FOO": "FOO"}),
        # Options matching the backend name update the defaults.
        ({"dummy_foo": "bar"}, {"FOO": "bar"}),
    ],
)
def test_utils_get_backend_instance(options, expected):
    """Test the `get_backend_instance` utility should return the expected result."""

    class DummyTestBackendSettings(BaseModel):
        """Represent a dummy backend setting."""

        FOO: str = "FOO"

    class DummyTestBackend:
        """Represent a dummy backend instance."""

        name = "dummy"
        settings_class = DummyTestBackendSettings

        def __init__(self, settings):
            self.settings = settings

    backend = ralph_utils.get_backend_instance(DummyTestBackend, options)
    assert isinstance(backend, DummyTestBackend)
    assert backend.settings.dict() == expected


@pytest.mark.parametrize("path,value", [(["foo", "bar"], "bar_value")])
def test_utils_get_dict_value_from_path_should_return_given_value(path, value):
    """Test the get_dict_value_from_path function should return the value when it's
    present.
    """
    dictionary = {"foo": {"bar": "bar_value"}}
    assert ralph_utils.get_dict_value_from_path(dictionary, path) == value


@pytest.mark.parametrize(
    "path",
    [
        ["foo", "bar", "baz", "qux"],
        ["foo", "not_bar"],
        ["not_foo", "bar"],
        None,
    ],
)
def test_utils_get_dict_value_from_path_should_return_none_when_value_does_not_exists(
    path,
):
    """Test the get_dict_value_from_path function should return None if the value is
    not found.
    """
    dictionary = {"foo": {"bar": "bar_value"}}
    assert ralph_utils.get_dict_value_from_path(dictionary, path) is None


def test_utils_set_dict_value_from_path_creating_new_fields():
    """Test when the fields are not present, set_dict_value_from_path should add
    them.
    """
    dictionary = {}
    ralph_utils.set_dict_value_from_path(dictionary, ["foo", "bar"], "baz")
    assert dictionary == {"foo": {"bar": "baz"}}


def test_utils_set_dict_value_from_path_updating_fields():
    """Test when the fields are present, set_dict_value_from_path should update
    them.
    """
    dictionary = {"foo": {"bar": "bar_value"}}
    ralph_utils.set_dict_value_from_path(dictionary, ["foo", "bar"], "baz")
    assert dictionary == {"foo": {"bar": "baz"}}



"""Ralph tracking logs filters."""

from typing import Any, Union

from .exceptions import EventKeyError


def anonymous(event: dict) -> Union[dict, Any]:
    """Remove anonymous events.

    Args:
        event (dict): Event to filter.

    Returns:
        event (dict): When event is not anonymous.
        None: Otherwise.

    Raises:
        EventKeyError: When the event does not contain the `username` key.

    """
    if "username" not in event:
        raise EventKeyError("Cannot filter anonymous event without 'username' key.")
    if not event.get("username", ""):
        return None

    return event



"""Base xAPI model configuration."""

from pydantic import BaseModel, Extra


class BaseModelWithConfig(BaseModel):
    """Pydantic model for base configuration shared among all models."""

    class Config:  # noqa: D106
        extra = Extra.forbid
        min_anystr_length = 1


class BaseExtensionModelWithConfig(BaseModel):
    """Pydantic model for extension configuration shared among all models."""

    class Config:  # noqa: D106
        extra = Extra.allow
        min_anystr_length = 0



"""`TinCan Vocabulary` verbs definitions."""

import sys
from typing import Dict, Optional

from ...base.verbs import BaseXapiVerb
from ...constants import LANG_EN_US_DISPLAY

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class ViewedVerb(BaseXapiVerb):
    """Pydantic model for viewed `verb`.

    Attributes:
        id (str): Consists of the value `http://id.tincanapi.com/verb/viewed`.
        display (dict): Consists of the dictionary `{"en-US": "viewed"}`.
    """

    id: Literal[
        "http://id.tincanapi.com/verb/viewed"
    ] = "http://id.tincanapi.com/verb/viewed"
    display: Optional[Dict[Literal[LANG_EN_US_DISPLAY], Literal["viewed"]]]


class DownloadedVerb(BaseXapiVerb):
    """Pydantic model for downloaded `verb`.

    Attributes:
        id (str): Consists of the value `http://id.tincanapi.com/verb/downloaded`.
        display (dict): Consists of the dictionary `{"en-US": "downloaded"}`.
    """

    id: Literal[
        "http://id.tincanapi.com/verb/downloaded"
    ] = "http://id.tincanapi.com/verb/downloaded"
    display: Optional[Dict[Literal[LANG_EN_US_DISPLAY], Literal["downloaded"]]]


class UnregisteredVerb(BaseXapiVerb):
    """Pydantic model for unregistered `verb`.

    Attributes:
        id (str): Consists of the value `http://id.tincanapi.com/verb/unregistered`.
        display (dict): Consists of the dictionary `{"en-US": "unregistered"}`.
    """

    id: Literal[
        "http://id.tincanapi.com/verb/unregistered"
    ] = "http://id.tincanapi.com/verb/unregistered"
    display: Optional[Dict[Literal[LANG_EN_US_DISPLAY], Literal["unregistered"]]]



"""Navigational event field definition."""

from pydantic import constr

from ...base import AbstractBaseEventField


class NavigationalEventField(AbstractBaseEventField):
    """Pydantic model for navigational `event` field.

    Note: All navigational statements are emitted from the browser.

    Attributes:
        id (str): Consists of the edX ID of the sequence.
        old (int): For `seq_goto`, it consists of the index of the unit being jumped to.
            For `seq_next` and `seq_prev`, it consists of the index of the unit being
            navigated to.
        new (int): For `seq_goto`, it consists of the index of the unit being jumped
            from. For `seq_next` and `seq_prev`, it consists of the index of the unit
            being navigated away from.
    """

    id: constr(
        regex=(
            r"^block-v1:[^\/+]+(\/|\+)[^\/+]+(\/|\+)[^\/?]+type"
            r"@sequential\+block@[a-f0-9]{32}$"
        )
    )
    new: int
    old: int



"""Browser event model definitions."""

import sys
from typing import Union

from pydantic import AnyUrl, constr

from .base import BaseEdxModel

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class BaseBrowserModel(BaseEdxModel):
    """Pydantic model for core browser statements.

    This type of event is triggered on (XHR) POST/GET requests to the `/event` URL.

    Attributes:
        event_source (str): Consists of the value `browser`.
        page (AnyUrl): Consists of the URL (with hostname) of the visited page.
            Retrieved with:
                `window.location.href` from the JavaScript front-end.
        session (str): Consists of the md5 encrypted Django session key or an empty
            string.
    """

    event_source: Literal["browser"]
    page: AnyUrl
    session: Union[constr(regex=r"^[a-f0-9]{32}$"), Literal[""]]



"""Async LRS data backend for Ralph."""

import logging
from io import IOBase
from typing import AsyncIterator, Iterable, Optional, Union
from urllib.parse import ParseResult, parse_qs, urljoin, urlparse

from httpx import AsyncClient, HTTPError, HTTPStatusError, RequestError
from pydantic import AnyHttpUrl, PositiveInt, parse_obj_as

from ralph.backends.data.base import (
    AsyncWritable,
    BaseAsyncDataBackend,
    BaseOperationType,
    DataBackendStatus,
)
from ralph.backends.data.lrs import LRSDataBackendSettings, StatementResponse
from ralph.backends.lrs.base import LRSStatementsQuery
from ralph.exceptions import BackendException
from ralph.utils import iter_by_batch

logger = logging.getLogger(__name__)


class AsyncLRSDataBackend(
    BaseAsyncDataBackend[LRSDataBackendSettings, LRSStatementsQuery],
    AsyncWritable,
):
    """Asynchronous LRS data backend."""

    name = "async_lrs"
    default_operation_type = BaseOperationType.CREATE
    unsupported_operation_types = {
        BaseOperationType.APPEND,
        BaseOperationType.UPDATE,
        BaseOperationType.DELETE,
    }

    def __init__(self, settings: Optional[LRSDataBackendSettings] = None) -> None:
        """Instantiate the LRS HTTP (basic auth) backend client.

        Args:
            settings (LRSDataBackendSettings or None): The LRS data backend settings.
                If `settings` is `None`, a default settings instance is used instead.
        """
        super().__init__(settings)
        self.base_url = parse_obj_as(AnyHttpUrl, self.settings.BASE_URL)
        self.auth = (self.settings.USERNAME, self.settings.PASSWORD)
        self._client = None

    @property
    def client(self) -> AsyncClient:
        """Create a `httpx.AsyncClient` if it doesn't exist."""
        if not self._client:
            headers = self.settings.HEADERS.dict(by_alias=True)
            self._client = AsyncClient(auth=self.auth, headers=headers)
        return self._client

    async def status(self) -> DataBackendStatus:
        """HTTP backend check for server status."""
        status_url = urljoin(self.base_url, self.settings.STATUS_ENDPOINT)
        try:
            response = await self.client.get(status_url)
            response.raise_for_status()
        except RequestError:
            logger.error("Unable to request the server")
            return DataBackendStatus.AWAY
        except HTTPStatusError:
            logger.error("Response raised an HTTP status of 4xx or 5xx")
            return DataBackendStatus.ERROR

        return DataBackendStatus.OK

    async def read(  # noqa: PLR0913
        self,
        query: Optional[Union[str, LRSStatementsQuery]] = None,
        target: Optional[str] = None,
        chunk_size: Optional[int] = None,
        raw_output: bool = False,
        ignore_errors: bool = False,
        prefetch: Optional[PositiveInt] = None,
        max_statements: Optional[PositiveInt] = None,
    ) -> Union[AsyncIterator[bytes], AsyncIterator[dict]]:
        """Get statements from LRS `target` endpoint.

        The `read` method defined in the LRS specification returns `statements` array
        and `more` IRL. The latter is required when the returned `statement` list has
        been limited.

        Args:
            query (str, LRSStatementsQuery):  The query to select records to read.
            target (str): Endpoint from which to read data (e.g. `/statements`).
                If target is `None`, `/xAPI/statements` default endpoint is used.
            chunk_size (int or None): The number of records or bytes to read in one
                batch, depending on whether the records are dictionaries or bytes.
            raw_output (bool): Controls whether to yield bytes or dictionaries.
                If the records are dictionaries and `raw_output` is set to `True`, they
                are encoded as JSON.
                If the records are bytes and `raw_output` is set to `False`, they are
                decoded as JSON by line.
            ignore_errors (bool): If `True`, errors during the read operation
                are ignored and logged. If `False` (default), a `BackendException`
                is raised if an error occurs.
            prefetch (int): The number of records to prefetch (queue) while yielding.
                If `prefetch` is `None` it defaults to `1` - no records are prefetched.
            max_statements: The maximum number of statements to yield.
        """
        statements = super().read(
            query,
            target,
            chunk_size,
            raw_output,
            ignore_errors,
            prefetch,
            max_statements,
        )
        async for statement in statements:
            yield statement

    async def _read_dicts(
        self,
        query: LRSStatementsQuery,
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,  # noqa: ARG002
    ) -> AsyncIterator[dict]:
        """Method called by `self.read` yielding dictionaries. See `self.read`."""
        if not target:
            target = self.settings.STATEMENTS_ENDPOINT

        if query.limit:
            logger.warning(
                "The limit query parameter value is overwritten by the chunk_size "
                "parameter value."
            )

        query.limit = chunk_size

        # Create request URL
        target = ParseResult(
            scheme=urlparse(self.base_url).scheme,
            netloc=urlparse(self.base_url).netloc,
            path=target,
            query="",
            params="",
            fragment="",
        ).geturl()

        statements = self._fetch_statements(
            target=target,
            query_params=query.dict(exclude_none=True, exclude_unset=True),
        )

        # Iterate through results
        try:
            async for statement in statements:
                yield statement
        except HTTPError as error:
            msg = "Failed to fetch statements: %s"
            logger.error(msg, error)
            raise BackendException(msg % (error,)) from error

    async def write(  # noqa: PLR0913
        self,
        data: Union[IOBase, Iterable[bytes], Iterable[dict]],
        target: Optional[str] = None,
        chunk_size: Optional[int] = None,
        ignore_errors: bool = False,
        operation_type: Optional[BaseOperationType] = None,
        concurrency: Optional[int] = None,
    ) -> int:
        """Write `data` records to the `target` endpoint and return their count.

        Args:
            data (Iterable): The data to write.
            target (str or None): Endpoint in which to write data (e.g. `/statements`).
                If `target` is `None`, `/xAPI/statements` default endpoint is used.
            chunk_size (int or None): The number of records or bytes to write in one
                batch, depending on whether `data` contains dictionaries or bytes.
                If `chunk_size` is `None`, a default value is used instead.
            ignore_errors (bool): If `True`, errors during the write operation
                are ignored and logged. If `False` (default), a `BackendException`
                is raised if an error occurs.
            operation_type (BaseOperationType or None): The mode of the write operation.
                If `operation_type` is `None`, the `default_operation_type` is used
                instead. See `BaseOperationType`.
            concurrency (int): The number of chunks to write concurrently.
                If `None` it defaults to `1`.
        """
        return await super().write(
            data, target, chunk_size, ignore_errors, operation_type, concurrency
        )

    async def _write_dicts(  # noqa: PLR0913
        self,
        data: Iterable[dict],
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,
        operation_type: BaseOperationType,  # noqa: ARG002
    ) -> int:
        """Method called by `self.write` writing dictionaries. See `self.write`."""
        if not target:
            target = self.settings.STATEMENTS_ENDPOINT

        target = ParseResult(
            scheme=urlparse(self.base_url).scheme,
            netloc=urlparse(self.base_url).netloc,
            path=target,
            query="",
            params="",
            fragment="",
        ).geturl()

        logger.debug(
            "Start writing to the %s endpoint (chunk size: %s)", target, chunk_size
        )

        count = 0
        for chunk in iter_by_batch(data, chunk_size):
            count += await self._post_and_raise_for_status(target, chunk, ignore_errors)

        logger.debug("Posted %d statements", count)
        return count

    async def close(self) -> None:
        """Close the `httpx.AsyncClient`.

        Raise:
            BackendException: If a failure occurs during the close operation.
        """
        if not self._client:
            logger.warning("No backend client to close.")
            return

        await self.client.aclose()

    async def _fetch_statements(self, target, query_params: dict):
        """Fetch statements from a LRS."""
        while True:
            response = await self.client.get(target, params=query_params)
            response.raise_for_status()
            statements_response = StatementResponse(**response.json())
            statements = statements_response.statements
            if isinstance(statements, dict):
                statements = [statements]

            for statement in statements:
                yield statement

            if not statements_response.more:
                break

            query_params.update(parse_qs(urlparse(statements_response.more).query))

    async def _post_and_raise_for_status(self, target, chunk, ignore_errors):
        """POST chunk of statements to `target` and return the number of insertions."""
        try:
            request = await self.client.post(target, json=chunk)
            request.raise_for_status()
            return len(chunk)
        except HTTPError as error:
            msg = "Failed to post statements: %s"
            if ignore_errors:
                logger.warning(msg, error)
                return 0
            logger.error(msg, error)
            raise BackendException(msg % (error,)) from error



"""Open Response Assessment events model definitions."""

import sys
from typing import Union

from pydantic import Json

from ralph.models.edx.browser import BaseBrowserModel
from ralph.models.edx.server import BaseServerModel
from ralph.models.selector import selector

from .fields.events import (
    ORAAssessEventField,
    ORACreateSubmissionEventField,
    ORAGetPeerSubmissionEventField,
    ORAGetSubmissionForStaffGradingEventField,
    ORASaveSubmissionEventField,
    ORAStaffAssessEventField,
    ORAStudentTrainingAssessExampleEventField,
    ORASubmitFeedbackOnAssessmentsEventField,
    ORAUploadFileEventField,
)

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class ORAGetPeerSubmission(BaseServerModel):
    """Pydantic model for `openassessmentblock.get_peer_submission` statement.

    The server emits this statement when a response is delivered to a learner for
    evaluation.

    Attributes:
        event (dict): See ORAGetPeerSubmissionEventField.
        event_type (str): Consists of the value
            `openassessmentblock.get_peer_submission`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(
        event_source="server", event_type="openassessmentblock.get_peer_submission"
    )

    event: ORAGetPeerSubmissionEventField
    event_type: Literal["openassessmentblock.get_peer_submission"]
    page: Literal["x_module"]


class ORAGetSubmissionForStaffGrading(BaseServerModel):
    """Pydantic model for `openassessmentblock.get_submission_for_staff_grading`
    statement.

    The server emits this statement when a course team member retrieves a learner's
    response for grading.

    Attributes:
        event (dict): See ORAGetSubmissionForStaffGradingEventField.
        event_type (str): Consists of the value
            `openassessmentblock.get_submission_for_staff_grading`.
        page (str): Consists of the value `x_module`.
    """  # noqa: D205

    __selector__ = selector(
        event_source="server",
        event_type="openassessmentblock.get_submission_for_staff_grading",
    )

    event: ORAGetSubmissionForStaffGradingEventField
    event_type: Literal["openassessmentblock.get_submission_for_staff_grading"]
    page: Literal["x_module"]


class ORAPeerAssess(BaseServerModel):
    """Pydantic model for `openassessmentblock.peer_assess` statement.

    The server emits this statement when a learner submits an assessment of a
    peer's response.

    Attributes:
        event (dict): See ORAAssessEventField.
        event_type (str): Consists of the value `openassessmentblock.peer_assess`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(
        event_source="server", event_type="openassessmentblock.peer_assess"
    )

    event: ORAAssessEventField
    event_type: Literal["openassessmentblock.peer_assess"]
    page: Literal["x_module"]


class ORASelfAssess(BaseServerModel):
    """Pydantic model for `openassessmentblock.self_assess` statement.

    The server emits this statement when a learner submits a self-assessment.

    Attributes:
        event (dict): See ORAAssessEventField.
        event_type (str): Consists of the value `openassessmentblock.self_assess`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(
        event_source="server", event_type="openassessmentblock.self_assess"
    )

    event: ORAAssessEventField
    event_type: Literal["openassessmentblock.self_assess"]
    page: Literal["x_module"]


class ORAStaffAssess(BaseServerModel):
    """Pydantic model for `openassessmentblock.staff_assess` statement.

    The server emits this statement when a course team member submits an assessment
    of a learner's response.

    Attributes:
        event (dict): See ORAStaffAssessEventField.
        event_type (str): Consists of the value `openassessmentblock.staff_assess`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(
        event_source="server", event_type="openassessmentblock.staff_assess"
    )

    event: ORAStaffAssessEventField
    event_type: Literal["openassessmentblock.staff_assess"]
    page: Literal["x_module"]


class ORASubmitFeedbackOnAssessments(BaseServerModel):
    """Pydantic model for `openassessmentblock.submit_feedback_on_assessments`
    statement.

    The server emits this statement when a learner submits a suggestion, opinion or
    other feedback about the assessment process.

    Attributes:
        event (dict): See ORASubmitFeedbackOnAssessmentsEventField.
        event_type (str): Consists of the value
            `openassessmentblock.submit_feedback_on_assessments`.
        page (str): Consists of the value `x_module`.
    """  # noqa: D205

    __selector__ = selector(
        event_source="server",
        event_type="openassessmentblock.submit_feedback_on_assessments",
    )

    event: ORASubmitFeedbackOnAssessmentsEventField
    event_type: Literal["openassessmentblock.submit_feedback_on_assessments"]
    page: Literal["x_module"]


class ORACreateSubmission(BaseServerModel):
    """Pydantic model for `openassessmentblock.create_submission` statement.

    The server emits this statement when a learner submits a response, a peer
    assessment or a self assessment.

    Attributes:
        event (dict): See ORACreateSubmissionEventField.
        event_type (str): Consists of the value `openassessmentblock.create_submission`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(
        event_source="server", event_type="openassessmentblock.create_submission"
    )

    event: ORACreateSubmissionEventField
    event_type: Literal["openassessmentblock.create_submission"]
    page: Literal["x_module"]


class ORASaveSubmission(BaseServerModel):
    """Pydantic model for `openassessmentblock.save_submission` statement.

    The server emits this statement when the user clicks on the
    <kbd>Save your progress</kbd> button to save the current state of the
    response to an open assessment question.

    Attributes:
        event (str): See ORASaveSubmissionEventField.
        event_type (str): Consists of the value `openassessmentblock.save_submission`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(
        event_source="server", event_type="openassessmentblock.save_submission"
    )

    event: ORASaveSubmissionEventField
    event_type: Literal["openassessmentblock.save_submission"]
    page: Literal["x_module"]


class ORAStudentTrainingAssessExample(BaseServerModel):
    """Pydantic model for `openassessment.student_training_assess_example` statement.

    The server emits this event when a learner submits an assessment for an example
    response within a training step.

    Attributes:
        event (dict): See ORAStudentTrainingAssessExampleEventField.
        event_type (str): Consists of the value
            `openassessment.student_training_assess_example`.
        page (str): Consists of the value `x_module`.
    """

    __selector__ = selector(
        event_source="server",
        event_type="openassessment.student_training_assess_example",
    )

    event: ORAStudentTrainingAssessExampleEventField
    event_type: Literal["openassessment.student_training_assess_example"]
    page: Literal["x_module"]


class ORAUploadFile(BaseBrowserModel):
    """Pydantic model for `openassessment.upload_file` statement.

    The browser emits this statement when a learner successfully uploads an image,
    .pdf, or other file as part of a response.

    Attributes:
        event (dict): See ORAUploadFileEventField.
        event_type (str): Consists of the value `openassessment.upload_file`.
        name (str): Consists of the value `openassessment.upload_file`.
    """

    __selector__ = selector(
        event_source="browser", event_type="openassessment.upload_file"
    )

    event: Union[Json[ORAUploadFileEventField], ORAUploadFileEventField]
    event_type: Literal["openassessment.upload_file"]
    name: Literal["openassessment.upload_file"]



"""Constants for `AcrossX Profile` xAPI profile."""

# Activity extensions
ACTIVITY_EXTENSIONS_TYPE = "https://w3id.org/xapi/acrossx/extensions/type"

# Context extensions
CONTEXT_EXTENSION_SCHOOL_ID = "https://w3id.org/xapi/acrossx/extensions/school"
